//========================================================================
//
// OPRS.cc
//
//========================================================================

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include <stdlib.h>
#include <string.h>
#include <string.h>
#include <dlfcn.h>
#include <stdarg.h>
#include "gmem.h"
#include "SplashErrorCodes.h"
#include "SplashMath.h"
#include "SplashBitmap.h"
#include "SplashState.h"
#include "SplashPath.h"
#include "SplashXPath.h"
#include "SplashXPathScanner.h"
#include "SplashPattern.h"
#include "SplashScreen.h"
#include "SplashClip.h"
#include "SplashFont.h"
#include "SplashGlyphBitmap.h"
#include "OPRS.h"
#include "Splash.h"
#include "OPVPSplash.h"

//------------------------------------------------------------------------
// Splash
//------------------------------------------------------------------------

#define SPLASH(x) (rasterMode ? (splash->x) : (opvpSplash->x))

OPRS::OPRS() 
{
  opvpHandle = NULL;
  splash = NULL;
  opvpSplash = NULL;
  rasterMode = gFalse;
}

int OPRS::setBitmap(SplashBitmap *bitmapA) {
  if (splash != NULL) {
    delete splash;
  }
  splash = new Splash(bitmapA);
  rasterMode = gTrue;
  return 0;
}

OPRS::~OPRS() {
  if (splash != NULL) {
    delete splash;
    splash = NULL;
  }
  if (opvpSplash != NULL) {
    delete opvpSplash;
    opvpSplash = NULL;
  }
  if (opvpHandle != NULL) {
    if (opvp != NULL && opvp->RestoreGS) {
	/* restore all GS to avoid bugs in some drivers  */
	while (opvp->RestoreGS(opvpContext) == 0);
    }
    if (opvp != NULL && opvp->ClosePrinter != NULL) {
      opvp->ClosePrinter(opvpContext);
    }
    unloadVectorDriver();
  }
}

//------------------------------------------------------------------------
// state read
//------------------------------------------------------------------------


SplashPattern *OPRS::getStrokePattern() {
  return SPLASH(getStrokePattern());
}

SplashPattern *OPRS::getFillPattern() {
  return SPLASH(getFillPattern());
}

SplashScreen *OPRS::getScreen() {
  return SPLASH(getScreen());
}

SplashCoord OPRS::getLineWidth() {
  return SPLASH(getLineWidth());
}

int OPRS::getLineCap() {
  return SPLASH(getLineCap());
}

int OPRS::getLineJoin() {
  return SPLASH(getLineJoin());
}

SplashCoord OPRS::getMiterLimit() {
  return SPLASH(getMiterLimit());
}

SplashCoord OPRS::getFlatness() {
  return 1;
}

SplashCoord *OPRS::getLineDash() {
  return SPLASH(getLineDash());
}

int OPRS::getLineDashLength() {
  return SPLASH(getLineDashLength());
}

SplashCoord OPRS::getLineDashPhase() {
  return SPLASH(getLineDashPhase());
}

SplashClip *OPRS::getClip() {
  return SPLASH(getClip());
}

//------------------------------------------------------------------------
// state write
//------------------------------------------------------------------------

void OPRS::setStrokePattern(SplashPattern *strokePattern) {
    SPLASH(setStrokePattern(strokePattern));
}

void OPRS::setFillPattern(SplashPattern *fillPattern) {
    SPLASH(setFillPattern(fillPattern));
}

void OPRS::setScreen(SplashScreen *screen) {
    SPLASH(setScreen(screen));
}

void OPRS::setLineWidth(SplashCoord lineWidth) {
    SPLASH(setLineWidth(lineWidth));
}

void OPRS::setMiterLimit(SplashCoord miterLimit) {
    SPLASH(setMiterLimit(miterLimit));
}

void OPRS::setLineCap(int lineCap) {
    SPLASH(setLineCap(lineCap));
}

void OPRS::setLineJoin(int lineJoin) {
    SPLASH(setLineJoin(lineJoin));
}

void OPRS::setFlatness(SplashCoord flatness) {
/* ignore flatness */
}

void OPRS::setLineDash(SplashCoord *lineDash, int lineDashLength,
			 SplashCoord lineDashPhase) {
    SPLASH(setLineDash(lineDash,lineDashLength,lineDashPhase));
}

SplashError OPRS::clipToPath(SplashPath *path, GBool eo) {
    return SPLASH(clipToPath(path,eo));
}

//------------------------------------------------------------------------
// state save/restore
//------------------------------------------------------------------------

void OPRS::saveState() {
    SPLASH(saveState());
}

SplashError OPRS::restoreState() {
    SPLASH(restoreState());
    return splashOk;
}

//------------------------------------------------------------------------
// drawing operations
//------------------------------------------------------------------------

void OPRS::clear(SplashColor color) {
    SPLASH(clear(color));
}

SplashError OPRS::stroke(SplashPath *path) {
    return SPLASH(stroke(path));
}

SplashError OPRS::fill(SplashPath *path, GBool eo) {
    return SPLASH(fill(path,eo));
}

SplashError OPRS::fillChar(SplashCoord x, SplashCoord y,
			     int c, SplashFont *font, Unicode *u, 
			     double *fontMat) {
    if (rasterMode) {
	return splash->fillChar(x,y,c,font,u);
    } else {
	return opvpSplash->fillChar(x,y,c,font,u,fontMat);
    }
}

SplashError OPRS::fillGlyph(SplashCoord x, SplashCoord y,
			      SplashGlyphBitmap *glyph) {
    return SPLASH(fillGlyph(x,y,glyph));
}

SplashError OPRS::fillImageMask(SplashImageMaskSource src, void *srcData,
				  int w, int h, SplashCoord *mat) {
    return SPLASH(fillImageMask(src,srcData,w,h,mat));
}

SplashError OPRS::drawImage(SplashImageSource src, void *srcData,
			      SplashColorMode srcMode,
			      int w, int h, SplashCoord *mat,
			      GBool mask) {
    if (rasterMode) {
	return splash->drawImage(src,srcData,srcMode,w,h,mat);
    } else {
	return opvpSplash->drawImage(src,srcData,srcMode,w,h,mat,mask);
    }
}

char * OPRS::allocString(char **destin, const char *source)
{
    if (!destin) return NULL;

    if (*destin) {
	if (source) {
	    *destin = (char *)realloc(*destin, strlen(source)+1);
	} else {
	    free(*destin);
	    *destin = NULL;
	}
    } else {
	if (source) {
	    *destin = (char *)malloc(strlen(source)+1);
	}
    }
    if (*destin && source) {
	if (*destin != source) {
	    strcpy(*destin, source);
	}
    }

    return *destin;
}


char ** OPRS::genDynamicLibName(const char *name)
{
    static char	*buff[5] = {NULL,NULL,NULL,NULL,NULL};
    char tbuff[OPVP_BUFF_SIZE];

    strcpy(tbuff, name);
    allocString(&(buff[0]), tbuff);
    strcat(tbuff, ".so");
    allocString(&(buff[1]), tbuff);
    strcpy(tbuff, name);
    strcat(tbuff, ".dll");
    allocString(&(buff[2]), tbuff);
    strcpy(tbuff, "lib");
    strcat(tbuff, name);
    strcat(tbuff, ".so");
    allocString(&(buff[3]), tbuff);
    buff[4] = NULL;

    return buff;
}

/*
 * initialize and load vector-driver
 */
int OPRS::init(const char *driverName, int outputFD,
  char *printerModel,int rasterModeA, int nOptions,
  char *optionKeys[], char *optionVals[])
{
    char **list = NULL;
    int	 i;
    void *h;
    int nApiEntry;

    list = genDynamicLibName(driverName);

    if (list) {
	i = 0;
	while (list[i]) {
	    if ((h = dlopen(list[i],RTLD_NOW))) {
		opvpOpenPrinter = (int (*)(int, char*, int
		  *, OPVP_api_procs**))dlsym(h,"OpenPrinter");
		opvpErrorno = (int *)dlsym(h,"errorno");
		if (opvpOpenPrinter && opvpErrorno) {
		    opvpHandle = h;
		    break;
		}
		opvpOpenPrinter = NULL;
		opvpErrorno = NULL;
	    }
	    i++;
	}
	for (i = 0;list[i] != NULL;i++) {
	    gfree(list[i]);
	    list[i] = 0;
	}
    }
    if (opvpHandle == NULL) {
      OPRS::error("Loading vector printer driver (%s) fail\n",driverName);
      return -1;
    }
    if ((opvpContext = (*opvpOpenPrinter)(outputFD,
	 printerModel,&nApiEntry,&opvp)) < 0) {
        OPRS::error("OpenPrinter fail\n",driverName);
	return -1;
    }
    if (opvp->NewPath == NULL) {
	/* force set rater mode */
	rasterModeA = gTrue;
    }
    rasterMode = rasterModeA;
    if (!rasterMode) {
	opvpSplash = new OPVPSplash(opvp,opvpContext,nOptions,
	  optionKeys, optionVals);
    }
    return 0;
}

/*
 * unload vector-driver
 */
int OPRS::unloadVectorDriver()
{
    if (opvpHandle != NULL) {
	dlclose(opvpHandle);
	opvpHandle = NULL;
	opvpOpenPrinter = NULL;
	opvpErrorno = NULL;
    }
    return 0;
}

int OPRS::OPVPStartJob(char *jobInfo)
{
    if (opvp->StartJob == NULL) {
      return 0;
    }
    return opvp->StartJob(opvpContext,jobInfo);
}

int OPRS::OPVPEndJob()
{
    if (opvp->EndJob == NULL) {
      return 0;
    }
    return opvp->EndJob(opvpContext);
}

int OPRS::OPVPStartDoc(char *docInfo)
{
    if (opvp->StartDoc == NULL) {
      return 0;
    }
    return opvp->StartDoc(opvpContext,docInfo);
}

int OPRS::OPVPEndDoc()
{
    if (opvp->EndDoc == NULL) {
      return 0;
    }
    return opvp->EndDoc(opvpContext);
}

int OPRS::OPVPStartPage(char *pageInfo, int rasterWidth)
{
    int r;

    if (opvp->StartPage != NULL) {
	if ((r = opvp->StartPage(opvpContext,pageInfo)) < 0) {
	  return r;
	}
    }
    if (opvp->SetROP != NULL) {
	opvp->SetROP(opvpContext,OPVP_ropPset);
    }
    if (rasterMode) {
	if (opvp->StartRaster == NULL) {
	    error("No StartRaster error in raster mode\n");
	    return -1;
	}
	if (opvp->SetCurrentPoint != NULL) {
	    OPVP_Fix x,y;

	    OPVP_f2Fix(0.0,x);
	    OPVP_f2Fix(0.0,y);
	    opvp->SetCurrentPoint(opvpContext,x,y);
	}
	opvp->StartRaster(opvpContext,rasterWidth);
    }
    return 0;
}

int OPRS::OPVPEndPage()
{
    int r;

    if (rasterMode) {
	if (opvp->EndRaster == NULL) {
	    error("No EndRaster error in raster mode\n");
	    return -1;
	}
	opvp->EndRaster(opvpContext);
    }
    if (opvp->EndPage != NULL) {
	if ((r = opvp->EndPage(opvpContext)) < 0) {
	    return r;
	}
    }
    return splashOk;
}

unsigned char *OPRS::getScanLineDataMono1(unsigned char *dst,
  unsigned char *bitmap, int rasterWidth)
{
    int n = (rasterWidth+7)/8;

    memcpy(dst,bitmap,n);
    return bitmap+n;
}

unsigned char *OPRS::getScanLineDataMono8(unsigned char *dst,
  unsigned char *bitmap, int rasterWidth)
{
    memcpy(dst,bitmap,rasterWidth);
    return bitmap+rasterWidth;
}

unsigned char *OPRS::getScanLineDataRGB8(unsigned char *dst,
  unsigned char *bitmap, int rasterWidth)
{
    int i;

    for (i = 0;i < rasterWidth;i++) {
#if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN
	bitmap++;
	*dst++ = *bitmap++;
	*dst++ = *bitmap++;
	*dst++ = *bitmap++;
#else
	dst[2] = *bitmap++;
	dst[1] = *bitmap++;
	dst[0] = *bitmap++;
	bitmap++;
	dst += 3;
#endif
    }
    return bitmap;
}

unsigned char *OPRS::getScanLineDataBGR8Packed(unsigned char *dst,
   unsigned char *bitmap, int rasterWidth)
{
    memcpy(dst,bitmap,rasterWidth*3);
    return bitmap+rasterWidth*3;
}

OPRS::GetScanLineDataFunT OPRS::getGetScanLineDataFun(SplashBitmap *bitmap)
{
    switch (bitmap->getMode()) {
    case splashModeMono1:
	return getScanLineDataMono1;
    case splashModeMono8:
	return getScanLineDataMono8;
    case splashModeRGB8:
	return getScanLineDataRGB8;
    case splashModeBGR8Packed:
	return getScanLineDataBGR8Packed;
    default:
	OPRS::error("Unknown bitmap mode\n");
	break;
    }
    return getScanLineDataMono8;
}

int OPRS::getRasterSize(SplashBitmap *bitmap)
{
    int rw = bitmap->getWidth();

    switch (bitmap->getMode()) {
    case splashModeMono1:
	return (rw+7)/8;
    case splashModeMono8:
	return rw;
    case splashModeRGB8:
    case splashModeBGR8Packed:
        return rw*3;
    default:
	OPRS::error("Unknown bitmap mode\n");
	break;
    }
    return 0;
}

GBool OPRS::checkAll1(unsigned char *bp, int n, int width, int mode)
{
    int lastbytemask = 0xff;
    int i;

    if (mode == splashModeMono1) {
	lastbytemask <<= (width & 0x7);
	lastbytemask &= 0xff;
    }
    for (i = 0;i < n-1;i++) {
	if (*bp++ != 0xff) return gFalse;
    }
    return (*bp & lastbytemask) == lastbytemask;
}

int OPRS::outSlice()
{
    if (rasterMode) {
	/* out bitmap */
	int rasterWidth;
	int nScanLines;
	int rasterSize;
	unsigned char *p;
	int i;
	SplashBitmap *bitmap;
	SplashColorPtr cp;
	unsigned char *bp;
	GetScanLineDataFunT fun;
	int mode;

	if (opvp->StartRaster == NULL || opvp->TransferRasterData == NULL
	   || opvp->EndRaster == NULL) {
	  OPRS::error("No raster supporting printer driver\n");
	  return -1;
	}

	bitmap = splash->getBitmap();
	rasterWidth = bitmap->getWidth();
	nScanLines = bitmap->getHeight();
	rasterSize = getRasterSize(bitmap);
	if ((bp = new unsigned char[rasterSize]) == NULL) {
	    OPRS::error("Not enough memory\n");
	    return -1;
	}
	cp = (bitmap->getDataPtr());
	p = reinterpret_cast<unsigned char *>(cp.rgb8);
	fun = getGetScanLineDataFun(bitmap);
	mode = bitmap->getMode();
	for (i = 0;i < nScanLines;i++) {
	    p = (*fun)(bp,p,rasterWidth);
	    if (opvp->SkipRaster != NULL
	        && checkAll1(bp,rasterSize,rasterWidth,mode)) {
	      /* all white, skip raster */
		opvp->SkipRaster(opvpContext,1);
	    } else {
		opvp->TransferRasterData(opvpContext,rasterSize,bp);
	    }
	}
	delete[] bp;
    }
    return 0;
}

int OPRS::setColorMode(int colorModeA)
{
    OPVP_ColorSpace cspace = OPVP_cspaceStandardRGB;

    if (opvp->GetColorSpace != NULL) opvp->GetColorSpace(opvpContext,&cspace);
    switch (cspace){
    case OPVP_cspaceBW:
	if (colorModeA != splashModeMono1) {
	  OPRS::error("not mono mode is specified on a monochrome printer\n");
	  return -1;
	}
	break;
    case OPVP_cspaceDeviceGray:
	break;
	if (colorModeA != splashModeMono1 && colorModeA != splashModeMono8) {
	    OPRS::error("colorMode is specified on not a color printer\n");
	    return -1;
	}
    default:
	/* rgb color */
	if (opvp->SetColorSpace != NULL) opvp->SetColorSpace(opvpContext,
	   OPVP_cspaceStandardRGB);
	break;
    }
    if (!rasterMode) {
	opvpSplash->setColorMode(colorModeA);
    }
    return 0;
}

SplashBitmap *OPRS::getBitmap()
{
    return SPLASH(getBitmap());
}

void OPRS::setDebugMode(GBool debugModeA)
{
    SPLASH(setDebugMode(debugModeA));
}

void OPRS::initGS(int colorMode, int w, int h, SplashColor paperColor)
{
  SplashColor color;

  if (!rasterMode && opvp->InitGS != NULL) {
    opvp->InitGS(opvpContext);
  }
  switch (colorMode) {
  case splashModeMono1: color.mono1 = 0; break;
  case splashModeMono8: color.mono8 = 0; break;
  case splashModeRGB8: color.rgb8 = splashMakeRGB8(0, 0, 0); break;
  case splashModeBGR8Packed: color.bgr8 = splashMakeBGR8(0, 0, 0); break;
  }
  if (!rasterMode) {
    opvpSplash->setStateBypass(gTrue);
  }
  SPLASH(setStrokePattern(new SplashSolidColor(color)));
  SPLASH(setFillPattern(new SplashSolidColor(color)));
  SPLASH(setLineCap(splashLineCapButt));
  SPLASH(setLineJoin(splashLineJoinMiter));
  SPLASH(setLineDash(NULL, 0, 0));
  SPLASH(setLineWidth(0));
  SPLASH(setMiterLimit(10));
  SPLASH(setFlatness(1));
  SPLASH(clipResetToRect(0,0,w-1,h-1));
  SPLASH(clear(paperColor));
  if (!rasterMode) {
    opvpSplash->setStateBypass(gFalse);
  }
}

void OPRS::error(char *msg, ...)
{
    va_list args;

    fprintf(stderr,"ERROR:OPRS:");
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
    fflush(stderr);
}

void OPRS::endPage()
{
    /* restore state */
    while (SPLASH(restoreState()) == splashOk);
    if (!rasterMode) {
	opvpSplash->endPage();
    }
}
