//========================================================================
//
// OPVPSplash.cc
//
//========================================================================

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "gmem.h"
#include "SplashErrorCodes.h"
#include "SplashMath.h"
#include "SplashBitmap.h"
#include "SplashState.h"
#include "SplashPath.h"
#include "SplashXPath.h"
#include "SplashXPathScanner.h"
#include "SplashPattern.h"
#include "SplashScreen.h"
#include "SplashClip.h"
#include "SplashFont.h"
#include "SplashGlyphBitmap.h"
#include "Splash.h"
#include "OPRS.h"
#include "OPVPSplash.h"

//------------------------------------------------------------------------
// OPVPSplash
//------------------------------------------------------------------------

OPVPSplash::OPVPSplash(OPVP_api_procs *opvpA, int printerContextA,
  int nOptions, char *optionKeys[], char *optionVals[])
{
  printerContext = printerContextA;
  opvp = new OPVPWrapper(opvpA);
  state = new SplashState(0,0);
  debugMode = gFalse;
  stateBypass = gFalse;
  clipPath = 0;
  if (getOption("OPVP_OLDLIPSDRIVER",nOptions,
     optionKeys,optionVals) != NULL) {
    oldLipsDriver = gTrue;
  } else {
    oldLipsDriver = gFalse;
  }
  if (getOption("OPVP_CLIPPATHNOTSAVED",nOptions,
     optionKeys,optionVals) != NULL) {
    clipPathNotSaved = gTrue;
  } else {
    clipPathNotSaved = gFalse;
  }
  if (getOption("OPVP_NOSHEARIMAGE",nOptions,
     optionKeys,optionVals) != NULL) {
    noShearImage = gTrue;
  } else {
    noShearImage = gFalse;
  }
  if (getOption("OPVP_NOLINESTYLE",nOptions,
     optionKeys,optionVals) != NULL) {
    noLineStyle = gTrue;
  } else {
    noLineStyle = gFalse;
  }
  if (opvpA->SetLineStyle == NULL || opvpA->SetLineDash == NULL
     || opvpA->SetLineDashOffset == NULL) {
    noLineStyle = gTrue;
  }
  if (getOption("OPVP_NOCLIPPATH",nOptions,
     optionKeys,optionVals) != NULL) {
    noClipPath = gTrue;
  } else {
    noClipPath = gFalse;
  }
  if (getOption("OPVP_NOBITMAPCHAR",nOptions,
     optionKeys,optionVals) != NULL) {
    noBitmapChar = gTrue;
  } else {
    noBitmapChar = gFalse;
  }
  if (opvpA->SetClipPath == NULL) {
    noClipPath = gTrue;
  }
  savedNoClipPath = noClipPath;
  /* dummy SaveGS for some drivers' BUG */
  /* some drivers can't restore the first saved GS */
  opvp->SaveGS(printerContext);
#ifdef OPTION_DEBUG
fprintf(stderr,"noClipPath=%d\n",noClipPath);
fprintf(stderr,"oldLipsDriver=%d\n",oldLipsDriver);
fprintf(stderr,"noLineStyle=%d\n",noLineStyle);
fprintf(stderr,"noShearImage=%d\n",noShearImage);
fprintf(stderr,"clipPathNotSaved=%d\n",clipPathNotSaved);
fprintf(stderr,"noBitmapChar=%d\n",noBitmapChar);
#endif
}

OPVPSplash::~OPVPSplash()
{
  while (state->next) {
    restoreState();
  }
  delete state;
  delete opvp;
}

//------------------------------------------------------------------------
// state read
//------------------------------------------------------------------------


SplashPattern *OPVPSplash::getStrokePattern() {
  return state->strokePattern;
}

SplashPattern *OPVPSplash::getFillPattern() {
  return state->fillPattern;
}

SplashScreen *OPVPSplash::getScreen() {
  return state->screen;
}

SplashCoord OPVPSplash::getLineWidth() {
  return state->lineWidth;
}

int OPVPSplash::getLineCap() {
  return state->lineCap;
}

int OPVPSplash::getLineJoin() {
  return state->lineJoin;
}

SplashCoord OPVPSplash::getMiterLimit() {
  return state->miterLimit;
}

SplashCoord OPVPSplash::getFlatness() {
  return state->flatness;
}

SplashCoord *OPVPSplash::getLineDash() {
  return state->lineDash;
}

int OPVPSplash::getLineDashLength() {
  return state->lineDashLength;
}

SplashCoord OPVPSplash::getLineDashPhase() {
  return state->lineDashPhase;
}

SplashClip *OPVPSplash::getClip() {
  return state->clip;
}

//------------------------------------------------------------------------
// state write
//------------------------------------------------------------------------

OPVP_ColorSpace OPVPSplash::getOPVPColorSpace()
{
  switch (colorMode) {
  case splashModeMono1:
    return OPVP_cspaceBW;
    break;
  case splashModeMono8:
    return OPVP_cspaceDeviceGray;
    break;
  case splashModeRGB8:
  case splashModeBGR8Packed:
  default:
    break;
  }
  return OPVP_cspaceStandardRGB;
}

void OPVPSplash::makeBrush(SplashPattern *pattern, OPVP_Brush *brush)
{
  brush->colorSpace = getOPVPColorSpace();
  brush->pbrush = NULL;
  brush->color[3] = -1;
  brush->xorg = brush->yorg = 0;
  if (pattern == NULL) {
    /* set default black color */
    brush->color[2] = 0;
    brush->color[1] = 0;
    brush->color[0] = 0;
  } else if (typeid(*pattern) == typeid(SplashHalftone)) {
    /* error */
    return;
  } else {
    /* solid color */
    SplashColor color;

    color = pattern->getColor(0,0);
    switch (colorMode) {
    case splashModeMono1:
      brush->color[2] = color.mono1;
      brush->color[1] = 0;
      brush->color[0] = 0;
      break;
    case splashModeMono8:
      brush->color[2] = color.mono8;
      brush->color[1] = 0;
      brush->color[0] = 0;
      break;
    case splashModeRGB8:
    case splashModeBGR8Packed:
      brush->color[2] = (color.rgb8 >> 16) & 0xff;
      brush->color[1] = (color.rgb8 >> 8) & 0xff;
      brush->color[0] = color.rgb8 & 0xff;
      break;
    default:
      OPRS::error("Unknown color mode\n");
      brush->color[2] = (color.rgb8 >> 16) & 0xff;
      brush->color[1] = (color.rgb8 >> 8) & 0xff;
      brush->color[0] = color.rgb8 & 0xff;
      break;
    }
  }
}

GBool OPVPSplash::equalPattern(SplashPattern *pat1, SplashPattern *pat2)
{
  SplashColor c1, c2;
  if (pat1 == NULL || pat2 == NULL) {
    return pat1 == pat2;
  }
  if (typeid(*pat1) != typeid(*pat2)) return gFalse;

  c1 = pat1->getColor(0,0);
  c2 = pat2->getColor(0,0);
  switch (colorMode) {
  case splashModeMono1:
    return c1.mono1 == c2.mono1;
    break;
  case splashModeMono8:
    return c1.mono8 == c2.mono8;
    break;
  case splashModeRGB8:
  case splashModeBGR8Packed:
    return c1.rgb8 == c2.rgb8;
    break;
  default:
    break;
  }
  return gTrue;
}

void OPVPSplash::setStrokePattern(SplashPattern *strokePattern) {
  OPVP_Brush brush;

  if (!stateBypass && equalPattern(strokePattern,state->strokePattern)) {
    delete strokePattern;
    return;
  }
  state->setStrokePattern(strokePattern);
  makeBrush(strokePattern,&brush);
  if (opvp->SetStrokeColor(printerContext,&brush) != 0) {
    OPRS::error("SetStrokeColor error\n");
    return;
  }
}

void OPVPSplash::setFillPattern(SplashPattern *fillPattern) {
  OPVP_Brush brush;

  if (!stateBypass && equalPattern(fillPattern,state->fillPattern)) {
    delete fillPattern;
    return;
  }
  state->setFillPattern(fillPattern);
  makeBrush(fillPattern,&brush);
  if (opvp->SetFillColor(printerContext,&brush) != 0) {
    OPRS::error("SetFillColor error\n");
    return;
  }
}

void OPVPSplash::setScreen(SplashScreen *screen) {
  state->setScreen(screen);
}

void OPVPSplash::setLineWidth(SplashCoord lineWidth) {
  if (stateBypass || state->lineWidth != lineWidth) {
      OPVP_Fix width;

      state->lineWidth = lineWidth;
      OPVP_f2Fix(lineWidth,width);
      if (opvp->SetLineWidth(printerContext,width) < 0) {
	OPRS::error("SetLineWidth error\n");
	return;
      }
  }
}

void OPVPSplash::setLineCap(int lineCap) {
  if (stateBypass || state->lineCap != lineCap) {
      OPVP_LineCap cap;

      state->lineCap = lineCap;
      switch (lineCap) {
      case splashLineCapButt:
	cap = OPVP_lineCapButt;
	break;
      case splashLineCapRound:
	cap = OPVP_lineCapRound;
	break;
      case splashLineCapProjecting:
	cap = OPVP_lineCapSquare;
	break;
      default:
	/* error */
	cap = OPVP_lineCapButt;
	break;
      }
      if (opvp->SetLineCap(printerContext,cap) < 0) {
	OPRS::error("SetLineCap error\n");
	return;
      }
  }
}

void OPVPSplash::setLineJoin(int lineJoin) {
  if (stateBypass || state->lineJoin != lineJoin) {
      OPVP_LineJoin join;

      state->lineJoin = lineJoin;
      switch (lineJoin) {
      case splashLineJoinMiter:
	join = OPVP_lineJoinMiter;
	break;
      case splashLineJoinRound:
	join = OPVP_lineJoinRound;
	break;
      case splashLineJoinBevel:
	join = OPVP_lineJoinBevel;
	break;
      default:
	/* error */
	join = OPVP_lineJoinMiter;
	break;
      }
      if (opvp->SetLineJoin(printerContext,join) < 0) {
	OPRS::error("SetLineJoin error\n");
	return;
      }
  }
}

void OPVPSplash::setMiterLimit(SplashCoord miterLimit) {
  if (stateBypass || state->miterLimit != miterLimit) {
      OPVP_Fix limit;

      state->miterLimit = miterLimit;
      if (oldLipsDriver) {
	/* for old driver for lips */
	/* miterLimit is length/2 */
	miterLimit = miterLimit*state->lineWidth*0.5;
      }
      OPVP_f2Fix(miterLimit,limit);
      if (opvp->SetMiterLimit(printerContext,limit) < 0) {
	OPRS::error("SetMiterLimit error\n");
	return;
      }
  }
}

void OPVPSplash::setFlatness(SplashCoord flatness) {
  if (flatness < 1) {
    state->flatness = 1;
  } else {
    state->flatness = flatness;
  }
}

void OPVPSplash::setLineDash(SplashCoord *lineDash, int lineDashLength,
			 SplashCoord lineDashPhase) {
  int i;
  OPVP_Fix *pdash;
  GBool equal;

  if (stateBypass || lineDash != state->lineDash) {
    if (lineDash == NULL || lineDashLength == 0) {
      if (opvp->SetLineStyle(printerContext,OPVP_lineStyleSolid) < 0) {
	OPRS::error("SetLineStyle error\n");
	return;
      }
      state->setLineDash(lineDash, lineDashLength, lineDashPhase);
      return;
    } else if (stateBypass || state->lineDash == NULL) {
      if (opvp->SetLineStyle(printerContext,OPVP_lineStyleDash) < 0) {
	OPRS::error("SetLineStyle error\n");
	return;
      }
    }
  }
  if (lineDash == NULL || lineDashLength == 0) return;
  equal = (state->lineDash != NULL);
  pdash = new OPVP_Fix[lineDashLength];
  for (i = 0;i < lineDashLength;i++) {
    if (equal && lineDash[i] != state->lineDash[i]) equal = gFalse;
    OPVP_f2Fix(lineDash[i],pdash[i]);
  }
  if (!equal && opvp->SetLineDash(printerContext,pdash,lineDashLength) < 0) {
    OPRS::error("SetLineDash error\n");
    goto err;
  }
  if (stateBypass || lineDashPhase != state->lineDashPhase) {
    OPVP_Fix offset;

    OPVP_f2Fix(lineDashPhase,offset);
    if (opvp->SetLineDashOffset(printerContext,offset) < 0) {
      OPRS::error("SetLineDashOffset error\n");
      goto err;
    }
  }
  state->setLineDash(lineDash, lineDashLength, lineDashPhase);
err:
  delete[] pdash;
}

SplashError OPVPSplash::doClipPath(SplashPath *path, GBool eo)
{
  SplashError result;

  if (path->length > OPVP_MAX_CLIPPATH_LENGTH) {
    noClipPath = gTrue;
  } else {
    noClipPath = savedNoClipPath;
  }
  if (!noClipPath && path->length > 0) {
    /* when path->length == 0, no drawable arae, and no output 
       so, it isn't need to set ClipPath */
    if ((result = makePath(path)) != splashOk) {
      return result;
    }
    if (opvp->SetClipPath(printerContext,
	 eo ? OPVP_clipRuleEvenOdd : OPVP_clipRuleWinding) < 0) {
      OPRS::error("SetClipPath error\n");
      return splashErrOPVP;
    }
  }
  return splashOk;
}

SplashError OPVPSplash::makeRectanglePath(SplashCoord x0,
  SplashCoord y0, SplashCoord x1, SplashCoord y1, SplashPath **p)
{
  SplashError result;

  *p = new SplashPath();
  if ((result = (*p)->moveTo(x0,y0)) != splashOk) return result;
  if ((result = (*p)->lineTo(x1,y0)) != splashOk) return result;
  if ((result = (*p)->lineTo(x1,y1)) != splashOk) return result;
  if ((result = (*p)->lineTo(x0,y1)) != splashOk) return result;
  if ((result = (*p)->close()) != splashOk) return result;
  return splashOk;
}

void OPVPSplash::clipResetToRect(SplashCoord x0, SplashCoord y0,
			     SplashCoord x1, SplashCoord y1) {
  SplashPath *p;
  OPVPClipPath *cp;

  while ((cp = OPVPClipPath::pop()) != NULL) delete cp;
  if (clipPath != 0) {
    delete clipPath;
    clipPath = 0;
  }

  if (makeRectanglePath(x0,y0,x1,y1,&p) != splashOk) return;

  if (doClipPath(p, gTrue) != splashOk) return;
  clipPath = new OPVPClipPath(p,gTrue);
  state->clip->resetToRect(x0, y0, x1, y1);
}

GBool OPVPSplash::isRectanglePath(SplashPath *path,
  SplashCoord *xMin, SplashCoord *yMin, SplashCoord *xMax, SplashCoord *yMax)
{
  if (path->length != 5
      || path->pts[0].x != path->pts[4].x
      || path->pts[0].y != path->pts[4].y
      || path->flags[0] != (splashPathFirst | splashPathClosed)
      || path->flags[1] != 0
      || path->flags[2] != 0
      || path->flags[3] != 0
      || path->flags[4] != (splashPathLast | splashPathClosed)) {
    return gFalse;
  }
  if (splashRound(path->pts[0].x) == splashRound(path->pts[1].x)) {
    if (splashRound(path->pts[1].y) != splashRound(path->pts[2].y)
      || splashRound(path->pts[2].x) != splashRound(path->pts[3].x)
      || splashRound(path->pts[3].y) != splashRound(path->pts[4].y)) {
      return gFalse;
    }
  } else if (splashRound(path->pts[0].y) == splashRound(path->pts[1].y)) {
    if (splashRound(path->pts[1].x) != splashRound(path->pts[2].x)
      || splashRound(path->pts[2].y) != splashRound(path->pts[3].y)
      || splashRound(path->pts[3].x) != splashRound(path->pts[4].x)) {
      return gFalse;
    }
  } else {
    return gFalse;
  }
  *xMin = path->pts[0].x;
  *yMin = path->pts[0].y;
  *xMax = path->pts[2].x;
  *yMax = path->pts[2].y;
  if (*xMin > *xMax) {
    SplashCoord t = *xMin;

    *xMin = *xMax;
    *xMax = t;
  }
  if (*yMin > *yMax) {
    SplashCoord t = *yMin;

    *yMin = *yMax;
    *yMax = t;
  }
  return gTrue;
}

SplashError OPVPSplash::clipToPath(SplashPath *path, GBool eo) {
  SplashError result;
  SplashCoord x0, y0, x1, y1;
  SplashCoord x2, y2, x3, y3;
  SplashClipResult clipResult;
  int xMin, yMin, xMax, yMax;

  if (path == 0) return splashErrBogusPath;
  if (path->length == 0) return splashOk;
  if (clipPath == 0) {
    /* no clip region exist */
    if ((result = state->clip->clipToPath(path, state->flatness, eo))
	 != splashOk) {
      return result;
    }
    path = path->copy();
  } else {
    SplashPath *oldPath = clipPath->getPath();
    if (isRectanglePath(path,&x0,&y0,&x1,&y1)) {
      if ((clipResult = state->clip->testRect(
         splashRound(x0), splashRound(y0), splashRound(x1), splashRound(y1)))
	 == splashClipAllOutside) {
	 /* no drawable area */
	if ((result = state->clip->clipToPath(path, state->flatness, eo))
	     != splashOk) {
	  return result;
	}
	path = new SplashPath();
      } else if (clipResult == splashClipPartial) {
	if (isRectanglePath(oldPath,&x2,&y2,&x3,&y3)) {
	  if ((result = state->clip->clipToPath(path, state->flatness, eo))
	       != splashOk) {
	    return result;
	  }
	  /* both rectangle */
	  if (x0 < x2) x0 = x2;
	  if (y0 < y2) y0 = y2;
	  if (x1 > x3) x1 = x3;
	  if (y1 > y3) y1 = y3;
	  if ((result = makeRectanglePath(x0,y0,x1,y1,&path)) != splashOk) {
	    return result;
	  }
	} else {
	  state->clip->getBBox(&xMin,&yMin,&xMax,&yMax);
	  if (splashRound(x0) <= xMin && splashRound(y0) <= yMin
	      && splashRound(x1) >= xMax && splashRound(y1)) {
		/* The old path is all inside the new path */
		/* We may ignore the new path */
	    return splashOk;
	  }
	  if ((result = state->clip->clipToPath(path, state->flatness, eo))
	       != splashOk) {
	    return result;
	  }
	  path = state->clip->makePath();
	}
      } else {
	/* splashClipAllInside */
	/* We may ignore the previous region. */
	if ((result = state->clip->clipToPath(path, state->flatness, eo))
	     != splashOk) {
	  return result;
	}
	path = path->copy();
      }
    } else {
      /* non rectangle path */

      SplashXPath *xpath = new SplashXPath(path, state->flatness, gFalse);

      xpath->sort();
      SplashXPathScanner *scanner = new SplashXPathScanner(xpath,eo);
      scanner->getBBox(&xMin,&yMin,&xMax,&yMax);
      delete scanner;
      delete xpath;
      if ((clipResult = state->clip->testRect(xMin,yMin,xMax,yMax))
         == splashClipAllOutside) {
	 /* no efect */
	 /* no drawable area */
	if ((result = state->clip->clipToPath(path, state->flatness, eo))
	     != splashOk) {
	  return result;
	}
	path = new SplashPath();
      } else if (clipResult == splashClipPartial) {
	SplashClip *nclip = new SplashClip(xMin,yMin,xMax,yMax);
	nclip->clipToPath(path,state->flatness,eo);
	state->clip->getBBox(&xMin,&yMin,&xMax,&yMax);
	if ((clipResult = nclip->testRect(xMin,yMin,xMax,yMax))
	   == splashClipAllOutside) {
	  /* no drawable area */
	  delete nclip;
	  if ((result = state->clip->clipToPath(path, state->flatness, eo))
		 != splashOk) {
	    return result;
	  }
	  path = new SplashPath();
	} else {
	    delete nclip;
	    if (clipResult == splashClipAllInside) {
	      /* The old path is all inside the new path */
	      /* We may ignore the new path */
	      return splashOk;
	    }
	    if ((result = state->clip->clipToPath(path, state->flatness, eo))
		 != splashOk) {
	      return result;
	    }
	    path = state->clip->makePath();
	}
      } else {
	/* splashClipAllInside */
	/* We may ignore the previous region. */
	if ((result = state->clip->clipToPath(path, state->flatness, eo))
	     != splashOk) {
	  return result;
	}
	path = path->copy();
      }
    }
  }
  if ((result = doClipPath(path,eo)) != splashOk) {
    delete path;
    return result;
  }
  if (clipPath != 0) delete clipPath;
  clipPath = new OPVPClipPath(path,eo);

  return splashOk;
}

//------------------------------------------------------------------------
// state save/restore
//------------------------------------------------------------------------

void OPVPSplash::saveState() {
  SplashState *newState;

  newState = state->copy();
  newState->next = state;
  state = newState;
  if (clipPath != 0) clipPath->push();
  if (opvp->SaveGS(printerContext) != 0) {
    OPRS::error("SaveGS error\n");
    return;
  }
}

SplashError OPVPSplash::restoreState() {
  SplashState *oldState;
  OPVPClipPath *oldClip;
  SplashPath *path;
  GBool saved = gFalse;

  if (!state->next) {
    return splashErrNoSave;
  }
  oldState = state;
  state = state->next;
  delete oldState;
  if (opvp->RestoreGS(printerContext) != 0) {
    OPRS::error("RestoreGS error\n");
    return splashErrOPVP;
  }
  oldClip = clipPath;
  if (clipPath != 0) {
    saved = clipPath->getSaved();
    delete clipPath;
    clipPath = 0;
  }
  clipPath = OPVPClipPath::pop();
  if (clipPath != 0) {
    path = clipPath->getPath();
    if (path->length > OPVP_MAX_CLIPPATH_LENGTH) {
      noClipPath = gTrue;
    } else {
      noClipPath = savedNoClipPath;
    }
  } else {
    noClipPath = savedNoClipPath;
  }
  if (clipPathNotSaved && !noClipPath) {
    if (clipPath != 0) {
      if (!saved) {
	  SplashError result;

	  if ((result = doClipPath(clipPath->getPath(), clipPath->getEo()))
		!= splashOk) return result;
      }
    } else if (oldClip != 0) {
      if (opvp->ResetClipPath(printerContext) != 0) {
	OPRS::error("ResetClipPath error\n");
	return splashErrOPVP;
      }
    }
  }
  return splashOk;
}

//------------------------------------------------------------------------
// drawing operations
//------------------------------------------------------------------------

void OPVPSplash::clear(SplashColor color) {
}

SplashError OPVPSplash::makePath(SplashPath *path)
{
  int i,j;
  OPVP_Fix x,y;

  if (opvp->NewPath(printerContext) < 0) {
    OPRS::error("NewPath error\n");
    return splashErrOPVP;
  }
  for (i = 0;i < path->length;i = j) {
    int curve = 0; 
    int n;
    OPVP_Point *points;
    int k;

    if ((path->flags[i] & splashPathFirst) != 0) {
      /* first point of a subpath */
      if ((path->flags[i] & splashPathLast) == 0
          || (path->flags[i] & splashPathClosed) != 0) {
	OPVP_f2Fix((path->pts[i].x),(x));
	OPVP_f2Fix((path->pts[i].y),(y));
	if (opvp->SetCurrentPoint(printerContext,x,y) < 0) {
	  OPRS::error("SetCurrentPoint error\n");
	  return splashErrOPVP;
	}
      }
      j = i+1;
      continue;
    }
    if (i+2 < path->length && path->flags[i] == splashPathCurve) {
      /* curve */
      curve = 1;
      for (j = i;j+2 < path->length
	 && path->flags[j] == splashPathCurve;j += 3);
    } else {
      curve = 0;
      for (j = i;j < path->length
	 && (path->flags[j] & splashPathCurve) == 0
	 && (path->flags[j] & splashPathFirst) == 0;j++);
    }

    n = j-i;
    points = new (OPVP_Point)[n];
    /* copy points */
    for (k = i; k < j;k++) {
      OPVP_f2Fix((path->pts[k].x),(points[k-i].x));
      OPVP_f2Fix((path->pts[k].y),(points[k-i].y));
    }

    if (curve) {
      /* curve */
      if (opvp->BezierPath(printerContext,n,points) < 0) {
	OPRS::error("BezierPath error\n");
	return splashErrOPVP;
      }
    } else {
      /* line */
      GBool closed = (path->flags[j-1] & splashPathClosed) != 0;

      if (closed) {
	/* if path is closed, 
	   the first point and the last point is identical.
	   So, the last point is not added. */

	if (opvp->LinePath(printerContext,OPVP_PathClose,
	      n-1,points) < 0) {
	  if (n-1 == 0) {
	    /* But some driver can't accept zero point */
	    /* So, retry with one point */
	    if (opvp->LinePath(printerContext,OPVP_PathClose,
		  n,points) < 0) {
	      OPRS::error("LinePath error\n");
	      return splashErrOPVP;
	    }
	  } else {
	    OPRS::error("LinePath error\n");
	    return splashErrOPVP;
	  }
	}
      } else {
	if (opvp->LinePath(printerContext, OPVP_PathOpen,
	      n,points) < 0) {
	  OPRS::error("LinePath error\n");
	  return splashErrOPVP;
	}
      }
    }
    delete[] points;
  }
  if (opvp->EndPath(printerContext) < 0) {
    OPRS::error("EndPath error\n");
    return splashErrOPVP;
  }
  return splashOk;
}

SplashXPath *OPVPSplash::makeDashedPath(SplashXPath *xPath) {
  SplashXPath *dPath;
  GBool lineDashStartOn, lineDashOn;
  GBool atSegStart, atSegEnd, atDashStart, atDashEnd;
  int lineDashStartIdx, lineDashIdx, subpathStart;
  SplashCoord lineDashTotal, lineDashStartPhase, lineDashDist;
  int segIdx;
  SplashXPathSeg *seg;
  SplashCoord sx0, sy0, sx1, sy1, ax0, ay0, ax1, ay1, dist;
  int i;

  dPath = new SplashXPath();

  lineDashTotal = 0;
  for (i = 0; i < state->lineDashLength; ++i) {
    lineDashTotal += state->lineDash[i];
  }
  lineDashStartPhase = state->lineDashPhase;
  i = splashFloor(lineDashStartPhase / lineDashTotal);
  lineDashStartPhase -= i * lineDashTotal;
  lineDashStartOn = gTrue;
  lineDashStartIdx = 0;
  while (lineDashStartPhase >= state->lineDash[lineDashStartIdx]) {
    lineDashStartOn = !lineDashStartOn;
    lineDashStartPhase -= state->lineDash[lineDashStartIdx];
    ++lineDashStartIdx;
  }

  segIdx = 0;
  seg = xPath->segs;
  sx0 = seg->x0;
  sy0 = seg->y0;
  sx1 = seg->x1;
  sy1 = seg->y1;
  dist = splashDist(sx0, sy0, sx1, sy1);
  lineDashOn = lineDashStartOn;
  lineDashIdx = lineDashStartIdx;
  lineDashDist = state->lineDash[lineDashIdx] - lineDashStartPhase;
  atSegStart = gTrue;
  atDashStart = gTrue;
  subpathStart = dPath->length;

  while (segIdx < xPath->length) {

    ax0 = sx0;
    ay0 = sy0;
    if (dist <= lineDashDist) {
      ax1 = sx1;
      ay1 = sy1;
      lineDashDist -= dist;
      dist = 0;
      atSegEnd = gTrue;
      atDashEnd = lineDashDist == 0 || (seg->flags & splashXPathLast);
    } else {
      ax1 = sx0 + (lineDashDist / dist) * (sx1 - sx0);
      ay1 = sy0 + (lineDashDist / dist) * (sy1 - sy0);
      sx0 = ax1;
      sy0 = ay1;
      dist -= lineDashDist;
      lineDashDist = 0;
      atSegEnd = gFalse;
      atDashEnd = gTrue;
    }

    if (lineDashOn) {
      dPath->addSegment(ax0, ay0, ax1, ay1,
			atDashStart, atDashEnd,
			atDashStart, atDashEnd);
      // end of closed subpath
      if (atSegEnd &&
	  (seg->flags & splashXPathLast) &&
	  !(seg->flags & splashXPathEnd1)) {
	dPath->segs[subpathStart].flags &= ~splashXPathEnd0;
	dPath->segs[dPath->length - 1].flags &= ~splashXPathEnd1;
      }
    }

    if (atDashEnd) {
      lineDashOn = !lineDashOn;
      if (++lineDashIdx == state->lineDashLength) {
	lineDashIdx = 0;
      }
      lineDashDist = state->lineDash[lineDashIdx];
      atDashStart = gTrue;
    } else {
      atDashStart = gFalse;
    }
    if (atSegEnd) {
      if (++segIdx < xPath->length) {
	++seg;
	sx0 = seg->x0;
	sy0 = seg->y0;
	sx1 = seg->x1;
	sy1 = seg->y1;
	dist = splashDist(sx0, sy0, sx1, sy1);
	if (seg->flags & splashXPathFirst) {
	  lineDashOn = lineDashStartOn;
	  lineDashIdx = lineDashStartIdx;
	  lineDashDist = state->lineDash[lineDashIdx] - lineDashStartPhase;
	  atDashStart = gTrue;
	  subpathStart = dPath->length;
	}
      }
      atSegStart = gTrue;
    } else {
      atSegStart = gFalse;
    }
  }

  return dPath;
}

void OPVPSplash::strokeNarrow(SplashXPath *xPath) {
  SplashXPathSeg *seg;
  int x0, x1, x2, x3, y0, y1, x, y, t;
  SplashCoord dx, dy, dxdy;
  SplashClipResult clipRes;
  int i;

  for (i = 0, seg = xPath->segs; i < xPath->length; ++i, ++seg) {

    x0 = splashFloor(seg->x0);
    x1 = splashFloor(seg->x1);
    y0 = splashFloor(seg->y0);
    y1 = splashFloor(seg->y1);

    // horizontal segment
    if (y0 == y1) {
      if (x0 > x1) {
	t = x0; x0 = x1; x1 = t;
      }
      if ((clipRes = state->clip->testSpan(x0, x1, y0))
	  != splashClipAllOutside) {
	drawSpan(x0, x1, y0, clipRes == splashClipAllInside);
      }

    // segment with |dx| > |dy|
    } else if (splashAbs(seg->dxdy) > 1) {
      dx = seg->x1 - seg->x0;
      dy = seg->y1 - seg->y0;
      dxdy = seg->dxdy;
      if (y0 > y1) {
	t = y0; y0 = y1; y1 = t;
	t = x0; x0 = x1; x1 = t;
	dx = -dx;
	dy = -dy;
      }
      if ((clipRes = state->clip->testRect(x0 <= x1 ? x0 : x1, y0,
					   x0 <= x1 ? x1 : x0, y1))
	  != splashClipAllOutside) {
	if (dx > 0) {
	  x2 = x0;
	  for (y = y0; y < y1; ++y) {
	    x3 = splashFloor(seg->x0 + (y + 1 - seg->y0) * dxdy);
	    drawSpan(x2, x3 - 1, y, clipRes == splashClipAllInside);
	    x2 = x3;
	  }
	  drawSpan(x2, x1, y, clipRes == splashClipAllInside);
	} else {
	  x2 = x0;
	  for (y = y0; y < y1; ++y) {
	    x3 = splashFloor(seg->x0 + (y + 1 - seg->y0) * dxdy);
	    drawSpan(x3 + 1, x2, y, clipRes == splashClipAllInside);
	    x2 = x3;
	  }
	  drawSpan(x1, x2, y, clipRes == splashClipAllInside);
	}
      }

    // segment with |dy| > |dx|
    } else {
      dxdy = seg->dxdy;
      if (y0 > y1) {
	t = y0; y0 = y1; y1 = t;
      }
      if ((clipRes = state->clip->testRect(x0 <= x1 ? x0 : x1, y0,
					   x0 <= x1 ? x1 : x0, y1))
	  != splashClipAllOutside) {
	for (y = y0; y <= y1; ++y) {
	  x = splashFloor(seg->x0 + (y - seg->y0) * dxdy);
	  drawPixel(x, y, clipRes == splashClipAllInside);
	}
      }
    }
  }
}

/*
  Translate arc to Bezier Curve

  input start point (x0,y0) , center (cx, cy) and end point (x3, y3)
  return Bezier curve control points (rx1,ry1 and rx2, ry2)

  an angle should be less than eqaul 90 degree

*/
void OPVPSplash::arcToCurve(SplashCoord x0, SplashCoord y0,
  SplashCoord x3, SplashCoord y3,
  SplashCoord cx, SplashCoord cy, SplashCoord *rx1, SplashCoord *ry1,
  SplashCoord *rx2, SplashCoord *ry2)
{
#define ROTX(x,y) (x*rotcos-y*rotsin)*r+cx
#define ROTY(x,y) (x*rotsin+y*rotcos)*r+cy

  SplashCoord x1,y1,x2,y2;
  SplashCoord r;
  SplashCoord rotcos, rotsin;
  SplashCoord ox,oy,hx,hy,d;

  hx = (x0+x3)/2;
  hy = (y0+y3)/2;
  r = splashDist(x0,y0,cx,cy);
  d = splashDist(x0,y0,hx,hy);
  rotcos = (hx-cx)/d;
  rotsin = (hy-cy)/d;
  oy = (splashDist(x0,y0,x3,y3)/2)/r;
  ox = splashDist(hx,hy,cx,cy)/r;
  x1 = ((4-ox)/3);
  y1 = ((1-ox)*(3-ox)/(3*oy));
  x2 = x1;
  y2 = -y1;
  *rx1 = ROTX(x1,y1);
  *ry1 = ROTY(x1,y1);
  *rx2 = ROTX(x2,y2);
  *ry2 = ROTY(x2,y2);
#undef ROTX
#undef ROTY
}

void OPVPSplash::strokeWide(SplashXPath *xPath) {
  SplashXPathSeg *seg, *seg2;
  SplashPath *widePath;
  SplashCoord d, dx, dy, wdx, wdy, dxPrev, dyPrev, wdxPrev, wdyPrev;
  SplashCoord dotprod, miter;
  SplashCoord x0,y0,x1,y1,x2,y2,x3,y3;
  int i, j;

  dx = dy = wdx = wdy = 0; // make gcc happy
  dxPrev = dyPrev = wdxPrev = wdyPrev = 0; // make gcc happy

  for (i = 0, seg = xPath->segs; i < xPath->length; ++i, ++seg) {

    // save the deltas for the previous segment; if this is the first
    // segment on a subpath, compute the deltas for the last segment
    // on the subpath (which may be used to draw a line join)
    if (seg->flags & splashXPathFirst) {
      for (j = i + 1, seg2 = &xPath->segs[j]; j < xPath->length; ++j, ++seg2) {
	if (seg2->flags & splashXPathLast) {
	  d = splashDist(seg2->x0, seg2->y0, seg2->x1, seg2->y1);
	  if (d == 0) {
	    //~ not clear what the behavior should be for joins with d==0
	    dxPrev = 0;
	    dyPrev = 1;
	  } else {
	    d = 1 / d;
	    dxPrev = d * (seg2->x1 - seg2->x0);
	    dyPrev = d * (seg2->y1 - seg2->y0);
	  }
	  wdxPrev = 0.5 * state->lineWidth * dxPrev;
	  wdyPrev = 0.5 * state->lineWidth * dyPrev;
	  break;
	}
      }
    } else {
      dxPrev = dx;
      dyPrev = dy;
      wdxPrev = wdx;
      wdyPrev = wdy;
    }

    // compute deltas for this line segment
    d = splashDist(seg->x0, seg->y0, seg->x1, seg->y1);
    if (d == 0) {
      // we need to draw end caps on zero-length lines
      //~ not clear what the behavior should be for splashLineCapButt with d==0
      dx = 0;
      dy = 1;
    } else {
      d = 1 / d;
      dx = d * (seg->x1 - seg->x0);
      dy = d * (seg->y1 - seg->y0);
    }
    wdx = 0.5 * state->lineWidth * dx;
    wdy = 0.5 * state->lineWidth * dy;

    // initialize the path (which will be filled)
    widePath = new SplashPath();
    widePath->moveTo(seg->x0 - wdy, seg->y0 + wdx);

    // draw the start cap
    if (seg->flags & splashXPathEnd0) {
      switch (state->lineCap) {
      case splashLineCapButt:
	widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
	break;
      case splashLineCapRound:
	x0 = seg->x0 - wdy;
	y0 = seg->y0 + wdx;
	x3 = seg->x0 - wdx;
	y3 = seg->y0 - wdy;
	arcToCurve(x0, y0, x3, y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x2,y2,x1,y1,x3,y3);
	x0 = x3;
	y0 = y3;
	x3 = seg->x0 + wdy;
	y3 = seg->y0 - wdx;
	arcToCurve(x0,y0,x3,y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x2,y2,x1,y1,x3,y3);
	break;
      case splashLineCapProjecting:
	widePath->lineTo(seg->x0 - wdx - wdy, seg->y0 + wdx - wdy);
	widePath->lineTo(seg->x0 - wdx + wdy, seg->y0 - wdx - wdy);
	widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
	break;
      }
    } else {
      widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
    }

    // draw the left side of the segment
    widePath->lineTo(seg->x1 + wdy, seg->y1 - wdx);

    // draw the end cap
    if (seg->flags & splashXPathEnd1) {
      switch (state->lineCap) {
      case splashLineCapButt:
	widePath->lineTo(seg->x1 - wdy, seg->y1 + wdx);
	break;
      case splashLineCapRound:
	x0 = seg->x1 + wdy;
	y0 = seg->y1 - wdx;
	x3 = seg->x1 + wdx;
	y3 = seg->y1 + wdy;
	arcToCurve(x0, y0, x3, y3,
	  seg->x1, seg->y1, &x1,&y1,&x2,&y2);
	widePath->curveTo(x2,y2,x1,y1,x3,y3);
	x0 = x3;
	y0 = y3;
	x3 = seg->x1 - wdy;
	y3 = seg->y1 + wdx;
	arcToCurve(x0,y0,x3,y3,
	  seg->x1, seg->y1, &x1,&y1,&x2,&y2);
	widePath->curveTo(x2,y2,x1,y1,x3,y3);
	break;
      case splashLineCapProjecting:
	widePath->lineTo(seg->x1 + wdx + wdy, seg->y1 - wdx + wdy);
	widePath->lineTo(seg->x1 + wdx - wdy, seg->y1 + wdx + wdy);
	widePath->lineTo(seg->x1 - wdy, seg->y1 + wdx);
	break;
      }
    } else {
      widePath->lineTo(seg->x1 - wdy, seg->y1 + wdx);
    }

    // draw the right side of the segment
    widePath->lineTo(seg->x0 - wdy, seg->y0 + wdx);

    // fill the segment
    fill(widePath, gTrue);
    delete widePath;

    // draw the line join
    if (!(seg->flags & splashXPathEnd0)) {
      widePath = NULL;
      switch (state->lineJoin) {
      case splashLineJoinMiter:
	dotprod = -(dx * dxPrev + dy * dyPrev);
	if (dotprod != 1) {
	  widePath = new SplashPath();
	  widePath->moveTo(seg->x0, seg->y0);
	  miter = 2 / (1 - dotprod);
	  if (splashSqrt(miter) <= state->miterLimit) {
	    miter = splashSqrt(miter - 1);
	    if (dy * dxPrev > dx * dyPrev) {
	      widePath->lineTo(seg->x0 + wdyPrev, seg->y0 - wdxPrev);
	      widePath->lineTo(seg->x0 + wdy - miter * wdx,
			       seg->y0 - wdx - miter * wdy);
	      widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
	    } else {
	      widePath->lineTo(seg->x0 - wdyPrev, seg->y0 + wdxPrev);
	      widePath->lineTo(seg->x0 - wdy - miter * wdx,
			       seg->y0 + wdx - miter * wdy);
	      widePath->lineTo(seg->x0 - wdy, seg->y0 + wdx);
	    }
	  } else {
	    if (dy * dxPrev > dx * dyPrev) {
	      widePath->lineTo(seg->x0 + wdyPrev, seg->y0 - wdxPrev);
	      widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
	    } else {
	      widePath->lineTo(seg->x0 - wdyPrev, seg->y0 + wdxPrev);
	      widePath->lineTo(seg->x0 - wdy, seg->y0 + wdx);
	    }
	  }
	}
	break;
      case splashLineJoinRound:
	widePath = new SplashPath();
	/* draw circle */
	widePath->moveTo(seg->x0 + wdy, seg->y0 - wdx);
	x0 = seg->x0 + wdy;
	y0 = seg->y0 - wdx;
	x3 = seg->x0 - wdx;
	y3 = seg->y0 - wdy;
	arcToCurve(x0, y0, x3, y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x1,y1,x2,y2,x3,y3);
	x0 = x3;
	y0 = y3;
	x3 = seg->x0 - wdy;
	y3 = seg->y0 + wdx;
	arcToCurve(x0, y0, x3, y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x1,y1,x2,y2,x3,y3);
	x0 = x3;
	y0 = y3;
	x3 = seg->x0 + wdx;
	y3 = seg->y0 + wdy;
	arcToCurve(x0, y0, x3, y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x1,y1,x2,y2,x3,y3);
	x0 = x3;
	y0 = y3;
	x3 = seg->x0 + wdy;
	y3 = seg->y0 - wdx;
	arcToCurve(x0, y0, x3, y3,
	  seg->x0, seg->y0, &x1,&y1,&x2,&y2);
	widePath->curveTo(x1,y1,x2,y2,x3,y3);
	break;
      case splashLineJoinBevel:
	widePath = new SplashPath();
	widePath->moveTo(seg->x0, seg->y0);
	if (dy * dxPrev > dx * dyPrev) {
	  widePath->lineTo(seg->x0 + wdyPrev, seg->y0 - wdxPrev);
	  widePath->lineTo(seg->x0 + wdy, seg->y0 - wdx);
	} else {
	  widePath->lineTo(seg->x0 - wdyPrev, seg->y0 + wdxPrev);
	  widePath->lineTo(seg->x0 - wdy, seg->y0 + wdx);
	}
	break;
      }
      if (widePath) {
	fill(widePath, gTrue);
	delete widePath;
      }
    }
  }
}

SplashError OPVPSplash::strokeByMyself(SplashPath *path)
{
  SplashXPath *xPath, *xPath2;
  SplashPattern *savedPattern;

  /* draw dashed line by myself */
  if (path->length == 0) {
    return splashOk;
  }
  xPath = new SplashXPath(path, state->flatness, gFalse);
  if (state->lineDash != NULL && state->lineDashLength > 0) {
    xPath2 = makeDashedPath(xPath);
    delete xPath;
    xPath = xPath2;
  }

  if (state->lineWidth <= 1) {
    strokeNarrow(xPath);
  } else {
    /* change fill pattern temprarily */
    savedPattern = state->fillPattern->copy();
    setFillPattern(state->strokePattern->copy());

    strokeWide(xPath);

    /* restore fill pattern */
    setFillPattern(savedPattern);
  }

  delete xPath;
  return splashOk;
}

SplashError OPVPSplash::stroke(SplashPath *path) {
  SplashError result;

  if (clipPath != 0 && clipPath->getPath()->length == 0) {
      return splashOk;
  }
  if ((state->lineDash != NULL
     && state->lineDashLength > 0 && noLineStyle)) {
    return strokeByMyself(path);
  }
  if (noClipPath) {
    int xMin, yMin, xMax, yMax;
    SplashClipResult clipResult;
    int fatOffset = splashCeil(state->lineWidth/2);
    int miterLimit = splashCeil(state->miterLimit/2);

    if (fatOffset < miterLimit) fatOffset = miterLimit;
    path->getBBox(&xMin,&yMin,&xMax,&yMax);
    xMin -= fatOffset;
    yMin -= fatOffset;
    xMax += fatOffset;
    yMax += fatOffset;
    clipResult = state->clip->testRect(xMin,yMin,xMax,yMax);
    if (clipResult == splashClipAllOutside) {
      /* not need to draw */
      return splashOk;
    } else if (clipResult == splashClipPartial) {
      return strokeByMyself(path);
    }
    /* splashClipAllInside */
    /* fall through */
  }
  if ((result = makePath(path)) != 0) return result;
  if (opvp->StrokePath(printerContext) < 0) {
    OPRS::error("StrokePath error\n");
    return splashErrOPVP;
  }
  return splashOk;
}

SplashError OPVPSplash::fillByMyself(SplashPath *path, GBool eo)
{
  SplashXPath *xPath;
  SplashXPathScanner *scanner;
  int xMinI, yMinI, xMaxI, yMaxI, x0, x1, y;
  SplashClipResult clipRes, clipRes2;

  if (path->length == 0) {
    return splashOk;
  }
  xPath = new SplashXPath(path, state->flatness, gTrue);
  xPath->sort();
  scanner = new SplashXPathScanner(xPath, eo);

  // get the min and max x and y values
  scanner->getBBox(&xMinI, &yMinI, &xMaxI, &yMaxI);

  // check clipping
  if ((clipRes = state->clip->testRect(xMinI, yMinI, xMaxI, yMaxI))
      != splashClipAllOutside) {
    SplashPattern *savedPattern;

    /* change stroke pattern temporarily */
    savedPattern = state->strokePattern->copy();
    setStrokePattern(state->fillPattern->copy());

    // draw the spans
    for (y = yMinI; y <= yMaxI; ++y) {
      while (scanner->getNextSpan(y, &x0, &x1)) {
	if (clipRes == splashClipAllInside) {
	  drawSpan(x0, x1, y, gTrue);
	} else {
	  clipRes2 = state->clip->testSpan(x0, x1, y);
	  drawSpan(x0, x1, y, clipRes2 == splashClipAllInside);
	}
      }
    }
    /* restore stroke pattern */
    setStrokePattern(savedPattern);
  }

  delete scanner;
  delete xPath;
  return splashOk;
}

SplashError OPVPSplash::fill(SplashPath *path, GBool eo) {
  SplashError result;
  OPVP_FillMode mode;

  if (clipPath != 0 && clipPath->getPath()->length == 0) {
      return splashOk;
  }
  if (noClipPath) {
    int xMin, yMin, xMax, yMax;
    SplashClipResult clipResult;

    path->getBBox(&xMin,&yMin,&xMax,&yMax);
    clipResult = state->clip->testRect(xMin,yMin,xMax,yMax);
    if (clipResult == splashClipAllOutside) {
      /* not need to draw */
      return splashOk;
    } else if (clipResult == splashClipPartial) {
      return fillByMyself(path,eo);
    }
    /* splashClipAllInside */
    /* fall through */
  }
  if ((result = makePath(path)) != 0) return result;
  mode = eo ? OPVP_fillModeEvenOdd : OPVP_fillModeWinding;
  if (opvp->SetFillMode(printerContext,mode) < 0) {
    OPRS::error("SetFillMode error\n");
    return splashErrOPVP;
  }
  if (opvp->FillPath(printerContext) < 0) {
    OPRS::error("FillPath error\n");
    return splashErrOPVP;
  }
  return splashOk;
}

SplashError OPVPSplash::fillGlyph(SplashCoord x, SplashCoord y,
  SplashGlyphBitmap *glyph)
{
  OPVP_Fix opvpx,opvpy;
  int opvpbytes;
  OPVP_Rectangle opvprect;
  int x0, y0;
  SplashError err;
  Guchar *bp;

  x0 = splashFloor(x)-glyph->x;
  y0 = splashFloor(y)-glyph->y;
  OPVP_i2Fix(0,opvprect.p0.x);
  OPVP_i2Fix(0,opvprect.p0.y);
  OPVP_i2Fix(glyph->w,opvprect.p1.x);
  OPVP_i2Fix(glyph->h,opvprect.p1.y);
  OPVP_i2Fix((x0),(opvpx));
  OPVP_i2Fix((y0),(opvpy));
  if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
    OPRS::error("SetCurrentPoint error\n");
  }

  if (oldLipsDriver && (((glyph->w+7)/8) & 3) != 0) {
    /* not 4bytes aligned, so make aligned */
    int i;
    int m = (glyph->w+7)/8;
    
    opvpbytes = (m+3)/4;
    opvpbytes *= 4;
    bp = (Guchar *)gmalloc(opvpbytes*glyph->h);
    for (i = 0;i < glyph->h;i++) {
      memcpy(bp+i*opvpbytes,glyph->data+i*m,m);
    }
  } else {
    bp = glyph->data;
    opvpbytes = (glyph->w+7)/8;
  }
  if (opvp->DrawImage(printerContext,glyph->w,glyph->h,1,OPVP_iformatRaw,
       opvprect, opvpbytes*glyph->h, (void *)bp) < 0) {
    OPRS::error("DrawImage error\n");
    err = splashErrOPVP;
  }
  if (bp != glyph->data) gfree(bp);

  return err;
}

SplashError OPVPSplash::fillChar(SplashCoord x, SplashCoord y,
			     int c, SplashFont *font,
			     Unicode *u, double *fontMat) {
  SplashError err = splashOk;
  SplashPath *path;
  double mx,my;

  path = font->getGlyphPath(c,u);
  if (!noBitmapChar) {
    mx = splashAbs(fontMat[0]);
    if (mx < splashAbs(fontMat[1])) {
	mx = splashAbs(fontMat[1]);
    }
    my = splashAbs(fontMat[3]);
    if (my < splashAbs(fontMat[2])) {
	my = splashAbs(fontMat[2]);
    }
    if (path == 0 || (mx*my < 2000)) {
      /* if a char is enough small, then out a char as a bitmask */
      SplashGlyphBitmap glyph;
      int x0, y0, xFrac, yFrac;
      SplashClipResult clipRes;

      x0 = splashFloor(x);
      xFrac = splashFloor((x - x0) * splashFontFraction);
      y0 = splashFloor(y);
      yFrac = splashFloor((y - y0) * splashFontFraction);
      if (font->getGlyph(c, xFrac, yFrac, &glyph, u)) {
	clipRes = state->clip->testRect(x0 - glyph.x,
			   y0 - glyph.y,
			   x0 - glyph.x + glyph.w - 1,
			   y0 - glyph.y + glyph.h - 1);

	if (!noClipPath || clipRes != splashClipPartial) {
	  if (path != 0) delete path;
	  if (clipRes != splashClipAllOutside) {
	    err = fillGlyph(x, y, &glyph);
	  }
	  if (glyph.freeData) {
	    gfree(glyph.data);
	  }
	  return err;
	} else {
	  if (glyph.freeData) {
	    gfree(glyph.data);
	  }
	}
      }
    }
    /* fall through and out a char as a path */
  }
  if (path == 0) {
    OPRS::error("FillPath error\n");
    err = splashErrOPVP;
    goto err0;
  }
  path->offset(x,y);
  err = fill(path,gFalse);
err0:
  if (path != 0) delete path;
  return err;
}

SplashError OPVPSplash::fillImageMaskFastWithCTM(SplashImageMaskSource src,
       void *srcData, int w, int h, int tx, int ty,SplashCoord *mat) {
  int i, j;
  OPVP_Fix opvpx,opvpy;
  int opvpbytes;
  int opvpcdepth;
  OPVP_Rectangle opvprect;
  OPVP_CTM opvpctm;
  Guchar *buf = 0, *bp;
  SplashError result = splashOk;

  opvpcdepth = 1;
  opvpbytes = (w+7)/8;
  /* align 4 */
  opvpbytes = (opvpbytes+3)/4;
  opvpbytes *= 4;
  buf = (Guchar *)gmalloc(opvpbytes*h);

  for (i = 0;i < h;i++) {
    int k;

    bp = buf+opvpbytes*i;
    for (j = 0;j < w;j += k) {
      SplashMono1 color;
      Guchar d;

      d = 0;
      for (k = 0;k < 8 && j+k < w;k++) {
	d <<= 1;
	(*src)(srcData, &color);
	if (color != 0) d |= 1;
      }
      d <<= 8-k;
      *bp++ = d;
    }
  }
  opvpctm.a = mat[0];
  opvpctm.b = mat[1];
  opvpctm.c = mat[2];
  opvpctm.d = mat[3];
  opvpctm.e = mat[4];
  opvpctm.f = mat[5];
  OPVP_i2Fix(0,opvprect.p0.x);
  OPVP_i2Fix(0,opvprect.p0.y);
  OPVP_i2Fix(1,opvprect.p1.x);
  OPVP_i2Fix(1,opvprect.p1.y);
  OPVP_i2Fix((tx),(opvpx));
  OPVP_i2Fix((ty),(opvpy));
  if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
    OPRS::error("SetCurrentPoint error\n");
  }

  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }
  if (opvp->DrawImage(printerContext,w,h,opvpcdepth,OPVP_iformatRaw,
       opvprect, opvpbytes*h, (void *)(buf)) < 0) {
    OPRS::error("DrawImage error\n");
    result = splashErrOPVP;
    goto err0;
  }
  err0:
  /* reset CTM */
  opvpctm.a = 1.0;
  opvpctm.b = 0.0;
  opvpctm.c = 0.0;
  opvpctm.d = 1.0;
  opvpctm.e = 0.0;
  opvpctm.f = 0.0;
  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }

  if (buf != 0) gfree(buf);
  return result;
}

SplashError OPVPSplash::fillImageMask(SplashImageMaskSource src, void *srcData,
				  int w, int h, SplashCoord *mat) {
  GBool rot;
  SplashCoord xScale, yScale, xShear, yShear;
  int tx, ty, scaledWidth, scaledHeight, xSign, ySign;
  int ulx, uly, llx, lly, urx, ury, lrx, lry;
  int ulx1, uly1, llx1, lly1, urx1, ury1, lrx1, lry1;
  int xMin, xMax, yMin, yMax;
  SplashClipResult clipRes;
  SplashMono1 *pixBuf;
  SplashMono1 *p;
#ifdef OLD_DRAW_IMAGE
  SplashClipResult clipRes2;
  int yp, yq, yt, yStep, lastYStep;
  int xp, xq, xt, xStep, xSrc;
  int k1, spanXMin, spanXMax, spanY;
  int pixAcc;
  int x1, y1, x2, y2;
  int n, m;
#endif
  int x, y;
  int i, j;
  SplashPattern *savedPattern;
  SplashCoord *savedLineDash = 0;
  int savedLineDashLength;
  SplashCoord savedLineDashPhase;
  SplashCoord savedLineWidth;

  if (debugMode) {
    printf("fillImageMask: w=%d h=%d mat=[%.2f %.2f %.2f %.2f %.2f %.2f]\n",
	   w, h, mat[0], mat[1], mat[2], mat[3], mat[4], mat[5]);
  }

  // check for singular matrix
  if (splashAbs(mat[0] * mat[3] - mat[1] * mat[2]) < 0.000001) {
    return splashErrSingularMatrix;
  }

  // compute scale, shear, rotation, translation parameters
  rot = splashAbs(mat[1]) > splashAbs(mat[0]);
  if (rot) {
    xScale = -mat[1];
    yScale = mat[2] - (mat[0] * mat[3]) / mat[1];
    xShear = -mat[3] / yScale;
    yShear = -mat[0] / mat[1];
  } else {
    xScale = mat[0];
    yScale = mat[3] - (mat[1] * mat[2]) / mat[0];
    xShear = mat[2] / yScale;
    yShear = mat[1] / mat[0];
  }
  tx = splashRound(mat[4]);
  ty = splashRound(mat[5]);
  scaledWidth = abs(splashRound(mat[4] + xScale) - tx) + 1;
  scaledHeight = abs(splashRound(mat[5] + yScale) - ty) + 1;
  xSign = (xScale < 0) ? -1 : 1;
  ySign = (yScale < 0) ? -1 : 1;

  // clipping
  ulx1 = 0;
  uly1 = 0;
  urx1 = xSign * (scaledWidth - 1);
  ury1 = splashRound(yShear * urx1);
  llx1 = splashRound(xShear * ySign * (scaledHeight - 1));
  lly1 = ySign * (scaledHeight - 1) + splashRound(yShear * llx1);
  lrx1 = xSign * (scaledWidth - 1) +
           splashRound(xShear * ySign * (scaledHeight - 1));
  lry1 = ySign * (scaledHeight - 1) + splashRound(yShear * lrx1);
  if (rot) {
    ulx = tx + uly1;    uly = ty - ulx1;
    urx = tx + ury1;    ury = ty - urx1;
    llx = tx + lly1;    lly = ty - llx1;
    lrx = tx + lry1;    lry = ty - lrx1;
  } else {
    ulx = tx + ulx1;    uly = ty + uly1;
    urx = tx + urx1;    ury = ty + ury1;
    llx = tx + llx1;    lly = ty + lly1;
    lrx = tx + lrx1;    lry = ty + lry1;
  }
  xMin = (ulx < urx) ? (ulx < llx) ? (ulx < lrx) ? ulx : lrx
                                   : (llx < lrx) ? llx : lrx
		     : (urx < llx) ? (urx < lrx) ? urx : lrx
                                   : (llx < lrx) ? llx : lrx;
  xMax = (ulx > urx) ? (ulx > llx) ? (ulx > lrx) ? ulx : lrx
                                   : (llx > lrx) ? llx : lrx
		     : (urx > llx) ? (urx > lrx) ? urx : lrx
                                   : (llx > lrx) ? llx : lrx;
  yMin = (uly < ury) ? (uly < lly) ? (uly < lry) ? uly : lry
                                   : (lly < lry) ? lly : lry
		     : (ury < lly) ? (ury < lry) ? ury : lry
                                   : (lly < lry) ? lly : lry;
  yMax = (uly > ury) ? (uly > lly) ? (uly > lry) ? uly : lry
                                   : (lly > lry) ? lly : lry
		     : (ury > lly) ? (ury > lry) ? ury : lry
                                   : (lly > lry) ? lly : lry;
  clipRes = state->clip->testRect(xMin, yMin, xMax, yMax);
  if (clipRes == splashClipAllOutside) return splashOk;

  if (!noClipPath || clipRes == splashClipAllInside) {
    if (!noShearImage) {
      if (fillImageMaskFastWithCTM(src,srcData,w,h,tx,ty,mat)
	  == splashOk) {
	return splashOk;
      }
    }
  }

#ifdef OLD_DRAW_IMAGE
  // compute Bresenham parameters for x and y scaling
  yp = h / scaledHeight;
  yq = h % scaledHeight;
  xp = w / scaledWidth;
  xq = w % scaledWidth;

  // allocate pixel buffer
  pixBuf = (SplashMono1 *)gmalloc((yp + 1) * w * sizeof(SplashMono1));

  /* change stroke pattern temprarily */
  savedPattern = state->strokePattern->copy();
  setStrokePattern(state->fillPattern->copy());

  /* change lins style temporarily */
  savedLineDashLength = state->lineDashLength;
  savedLineDashPhase = state->lineDashPhase;
  if (savedLineDashLength > 0 && state->lineDash != 0) {
    savedLineDash = new SplashCoord[savedLineDashLength];
    memcpy(savedLineDash, state->lineDash,
      savedLineDashLength*sizeof(SplashCoord));
  }
  setLineDash(0,0,0);
  savedLineWidth = state->lineWidth;
  setLineWidth(0.0);

  // init y scale Bresenham
  yt = 0;
  lastYStep = 1;

  for (y = 0; y < scaledHeight; ++y) {

    // y scale Bresenham
    yStep = yp;
    yt += yq;
    if (yt >= scaledHeight) {
      yt -= scaledHeight;
      ++yStep;
    }

    // read row(s) from image
    n = (yp > 0) ? yStep : lastYStep;
    if (n > 0) {
      p = pixBuf;
      for (i = 0; i < n; ++i) {
	for (j = 0; j < w; ++j) {
	  (*src)(srcData, p++);
	}
      }
    }
    lastYStep = yStep;

    // loop-invariant constants
    k1 = splashRound(xShear * ySign * y);

    // clipping test
    if (clipRes != splashClipAllInside &&
	!rot &&
	splashRound(yShear * k1) ==
	  splashRound(yShear * (xSign * (scaledWidth - 1) + k1))) {
      if (xSign > 0) {
	spanXMin = tx + k1;
	spanXMax = spanXMin + (scaledWidth - 1);
      } else {
	spanXMax = tx + k1;
	spanXMin = spanXMax - (scaledWidth - 1);
      }
      spanY = ty + ySign * y + splashRound(xShear * ySign * y);
      clipRes2 = state->clip->testSpan(spanXMin, spanXMax, spanY);
      if (clipRes2 == splashClipAllOutside) {
	continue;
      }
    } else {
      clipRes2 = clipRes;
    }

    // init x scale Bresenham
    xt = 0;
    xSrc = 0;

    for (x = 0; x < scaledWidth; ++x) {

      // x scale Bresenham
      xStep = xp;
      xt += xq;
      if (xt >= scaledWidth) {
	xt -= scaledWidth;
	++xStep;
      }

      // x shear
      x1 = xSign * x + k1;

      // y shear
      y1 = ySign * y + splashRound(yShear * x1);

      // rotation
      if (rot) {
	x2 = y1;
	y2 = -x1;
      } else {
	x2 = x1;
	y2 = y1;
      }

      // compute the alpha value for (x,y) after the x and y scaling
      // operations
      n = yStep > 0 ? yStep : 1;
      m = xStep > 0 ? xStep : 1;
      p = pixBuf + xSrc;
      pixAcc = 0;
      for (i = 0; i < n; ++i) {
	for (j = 0; j < m; ++j) {
	  pixAcc += *p++;
	}
	p += w - m;
      }

      // blend fill color with background
      if (pixAcc != 0) {
	drawPixel(tx + x2, ty + y2, clipRes2 == splashClipAllInside);
      }

      // x scale Bresenham
      xSrc += xStep;
    }
  }

  /* restore stroke pattern */
  setStrokePattern(savedPattern);
  /* restore line style */
  setLineDash(savedLineDash,savedLineDashLength,
    savedLineDashPhase);
  if (savedLineDash != 0) {
    delete[] savedLineDash;
  }
  setLineWidth(savedLineWidth);

  // free memory
  gfree(pixBuf);

  return splashOk;
#else
  SplashError result = splashOk;
  /* change stroke pattern temprarily */
  savedPattern = state->strokePattern->copy();
  setStrokePattern(state->fillPattern->copy());

  /* change lins style temporarily */
  savedLineDashLength = state->lineDashLength;
  savedLineDashPhase = state->lineDashPhase;
  if (savedLineDashLength > 0 && state->lineDash != 0) {
    savedLineDash = new SplashCoord[savedLineDashLength];
    memcpy(savedLineDash, state->lineDash,
      savedLineDashLength*sizeof(SplashCoord));
  }
  setLineDash(0,0,0);
  savedLineWidth = state->lineWidth;
  setLineWidth(0.0);

  /* calculate inverse matrix */
  SplashCoord imat[4];
  double det = mat[0] * mat[3] - mat[1] * mat[2];
  imat[0] = mat[3]/det;
  imat[1] = -mat[1]/det;
  imat[2] = -mat[2]/det;
  imat[3] = mat[0]/det;

  /* read source image */
  pixBuf = (SplashMono1 *)gmalloc(h * w * sizeof(SplashMono1));

  p = pixBuf;
  for (i = 0; i < h; ++i) {
    for (j = 0; j < w; ++j) {
      (*src)(srcData, p++);
    }
  }
  int width = xMax-xMin+1;
  int height = yMax-yMin+1;
  SplashClip *clip = state->clip->copy();

  if (w < scaledWidth || h < scaledHeight) {
    SplashPath cpath;

    cpath.moveTo(tx,ty);
    cpath.lineTo(mat[0]+tx,mat[1]+ty);
    cpath.lineTo(mat[0]+mat[2]+tx,mat[1]+mat[3]+ty);
    cpath.lineTo(mat[2]+tx,mat[3]+ty);
    clip->clipToPath(&cpath,1.0,gFalse);
  }
  for (y = 0;y < height;y++) {
    int dy = y+yMin-ty;
    int sx;
    GBool dmode = gFalse;

    for (x = 0;x < width;x++) {
      if (!clip->test(x+xMin,y+yMin)) {
	if (dmode) {
	  drawSpan(xMin+sx,xMin+x-1,yMin+y,gTrue);
	  dmode = gFalse;
	}
	continue;
      }
      int ox,oy;
      /* calculate original coordinate */
      int dx = x+xMin-tx;
      ox = (int)trunc((imat[0]*dx+imat[2]*dy)*w);
      oy = (int)trunc((imat[1]*dx+imat[3]*dy)*h);
      if (ox >= 0 && ox < w && oy >= 0 && oy < h) {
	GBool on = pixBuf[oy*w+ox] != 0;

	if (on && !dmode) {
	  dmode = gTrue;
	  sx = x;
	} else if (!on && dmode) {
	  drawSpan(xMin+sx,xMin+x-1,yMin+y,gTrue);
	  dmode = gFalse;
	}
      } else if (dmode) {
	drawSpan(xMin+sx,xMin+x-1,yMin+y,gTrue);
	dmode = gFalse;
      }
    }
    if (dmode) {
      drawSpan(xMin+sx,xMin+x-1,yMin+y,gTrue);
    }
  }
  delete clip;
  gfree(pixBuf);

  /* restore stroke pattern */
  setStrokePattern(savedPattern);
  /* restore line style */
  setLineDash(savedLineDash,savedLineDashLength,
    savedLineDashPhase);
  if (savedLineDash != 0) {
    delete[] savedLineDash;
  }
  setLineWidth(savedLineWidth);

  return result;

#endif
}

SplashError OPVPSplash::drawImageNotShear(SplashImageSource src,
                              void *srcData,
			      int w, int h,
			      int tx, int ty,
			      int scaledWidth, int scaledHeight,
			      int xSign, int ySign, GBool rot) {
  int i, j;
  OPVP_Fix opvpx,opvpy;
  int opvpbytes, linesize;
  int opvpcdepth;
  OPVP_Rectangle opvprect;
  OPVP_CTM opvpctm;
  SplashError result = splashOk;
  Guchar *buf = 0, *bp;
  float e,f;
  int hs,he, hstep;

  if (rot) {
    int t = h;

    h = w;
    w = t;
    t = scaledHeight;
    scaledHeight = scaledWidth;
    scaledWidth = t;
    if (xSign != ySign) {
	xSign = xSign >= 0 ? -1 : 1;
    } else {
	ySign = ySign >= 0 ? -1 : 1;
    }
  }
  OPVP_i2Fix(0,opvprect.p0.x);
  OPVP_i2Fix(0,opvprect.p0.y);
  OPVP_i2Fix(scaledWidth,opvprect.p1.x);
  OPVP_i2Fix(scaledHeight,opvprect.p1.y);

  if (xSign > 0) {
    OPVP_i2Fix((tx),(opvpx));
    e = tx;
  } else {
    OPVP_i2Fix((tx-scaledWidth),(opvpx));
    e = tx-scaledWidth;
  }
  if (ySign > 0) {
    OPVP_i2Fix((ty),(opvpy));
    f = ty;
  } else {
    OPVP_i2Fix((ty-scaledHeight),(opvpy));
    f = ty-scaledHeight;
  }
  if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
    OPRS::error("SetCurrentPoint error\n");
    return splashErrOPVP;
  }
  switch (colorMode) {
  case splashModeMono1:
  case splashModeMono8:
    opvpcdepth = 8;
    linesize = w;
    opvpbytes = (w+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeRGB8:
    opvpcdepth = 24;
    linesize = w*3;;
    opvpbytes = (w*3+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeBGR8Packed:
    OPRS::error("Image: no supported color mode\n");
    return splashErrOPVP;
    break;
  }
  if (ySign >= 0) {
    hstep = 1;
    hs = 0;
    he = h;
  } else {
    hstep = -1;
    hs = h-1;
    he = -1;
  }
  buf = (Guchar *)gmalloc(opvpbytes*h);
  switch (colorMode) {
  case splashModeMono1:
  case splashModeMono8:
    if (rot) {
      if (xSign >= 0) {
	for (i = 0;i < w;i++) {
	  for (j = hs;j != he;j += hstep) {
	    Guchar alpha;
	    SplashColor color;
	    bp = buf+i+j*opvpbytes;

	    (*src)(srcData, &color,&alpha);
	    *bp = color.mono8;
	  }
	}
      } else {
	for (i = 0;i < w;i++) {
	  for (j = hs;j != he;j += hstep) {
	    Guchar alpha;
	    SplashColor color;
	    bp = buf+linesize-1-i+j*opvpbytes;

	    (*src)(srcData, &color,&alpha);
	    *bp = color.mono8;
	  }
	}
      }
    } else {
      if (xSign >= 0) {
	for (i = hs;i != he;i += hstep) {
	  bp = buf+opvpbytes*i;
	  for (j = 0;j < w;j++) {
	    Guchar alpha;
	    SplashColor color;

	    (*src)(srcData, &color,&alpha);
	    *bp++ = color.mono8;
	  }
	}
      } else {
	for (i = hs;i != he;i += hstep) {
	  bp = buf+opvpbytes*i+linesize-1;
	  for (j = 0;j < w;j++) {
	    Guchar alpha;
	    SplashColor color;

	    (*src)(srcData, &color,&alpha);
	    *bp-- = color.mono8;
	  }
	}
      }
    }
    break;
  case splashModeRGB8:
    if (rot) {
      if (xSign >= 0) {
	for (i = 0;i < w;i++) {
	  for (j = hs;j != he;j += hstep) {
	    Guchar alpha;
	    SplashColor color;
	    bp = buf+i*3+j*opvpbytes;

	    (*src)(srcData, &color,&alpha);
	    bp[0] = splashRGB8R(color.rgb8);
	    bp[1] = splashRGB8G(color.rgb8);
	    bp[2] = splashRGB8B(color.rgb8);
	  }
	}
      } else {
	for (i = 0;i < w;i++) {
	  for (j = hs;j != he;j += hstep) {
	    Guchar alpha;
	    SplashColor color;
	    bp = buf+linesize-3-i*3+j*opvpbytes;

	    (*src)(srcData, &color,&alpha);
	    bp[0] = splashRGB8R(color.rgb8);
	    bp[1] = splashRGB8G(color.rgb8);
	    bp[2] = splashRGB8B(color.rgb8);
	  }
	}
      }
    } else {
      if (xSign >= 0) {
	for (i = hs;i != he;i += hstep) {
	  bp = buf+opvpbytes*i;
	  for (j = 0;j < w;j++) {
	    Guchar alpha;
	    SplashColor color;

	    (*src)(srcData, &color,&alpha);
	    *bp++ = splashRGB8R(color.rgb8);
	    *bp++ = splashRGB8G(color.rgb8);
	    *bp++ = splashRGB8B(color.rgb8);
	  }
	}
      } else {
	for (i = hs;i != he;i += hstep) {
	  bp = buf+opvpbytes*i+linesize-1;
	  for (j = 0;j < w;j++) {
	    Guchar alpha;
	    SplashColor color;

	    (*src)(srcData, &color,&alpha);
	    *bp-- = splashRGB8B(color.rgb8);
	    *bp-- = splashRGB8G(color.rgb8);
	    *bp-- = splashRGB8R(color.rgb8);
	  }
	}
      }
    }
    break;
  case splashModeBGR8Packed:
    OPRS::error("Image: no supported color mode\n");
    result = splashErrOPVP;
    goto err1;
    break;
  }

  /* canonlisp driver use CTM only, ignores currentPoint */
  /* So, set start point to CTM */
  opvpctm.a = 1.0;
  opvpctm.b = 0.0;
  opvpctm.c = 0.0;
  opvpctm.d = 1.0;
  opvpctm.e = e;
  opvpctm.f = f;
  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }

  if (opvp->DrawImage(printerContext,w,h,opvpcdepth,OPVP_iformatRaw,
       opvprect,opvpbytes*h, (void *)(buf)) < 0) {
    OPRS::error("DrawImage error\n");
    result = splashErrOPVP;
    goto err1;
  }
err1:
  /* reset CTM */
  opvpctm.e = 0.0;
  opvpctm.f = 0.0;
  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }

  if (buf != 0) gfree(buf);
  return result;
}

SplashError OPVPSplash::drawImageFastWithCTM(SplashImageSource src,
                              void *srcData,
			      int w, int h, int tx, int ty,
			      SplashCoord *mat) {
  int i, j;
  OPVP_Fix opvpx,opvpy;
  int opvpbytes;
  int opvpcdepth;
  OPVP_Rectangle opvprect;
  OPVP_CTM opvpctm;
  SplashError result = splashOk;
  Guchar *buf = 0, *bp;

  switch (colorMode) {
  case splashModeMono1:
  case splashModeMono8:
    opvpcdepth = 8;
    opvpbytes = (w+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeRGB8:
    opvpcdepth = 24;
    opvpbytes = (w*3+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeBGR8Packed:
    OPRS::error("Image: no supported color mode\n");
    return splashErrOPVP;
    break;
  }
  buf = (Guchar *)gmalloc(opvpbytes*h);

  switch (colorMode) {
  case splashModeMono1:
  case splashModeMono8:
    for (i = 0;i < h;i++) {
      bp = buf+opvpbytes*i;
      for (j = 0;j < w;j++) {
	Guchar alpha;
	SplashColor color;

	(*src)(srcData, &color,&alpha);
	*bp++ = color.mono8;
      }
    }
    break;
  case splashModeRGB8:
    for (i = 0;i < h;i++) {
      bp = buf+opvpbytes*i;
      for (j = 0;j < w;j++) {
	Guchar alpha;
	SplashColor color;

	(*src)(srcData, &color,&alpha);
	*bp++ = splashRGB8R(color.rgb8);
	*bp++ = splashRGB8G(color.rgb8);
	*bp++ = splashRGB8B(color.rgb8);
      }
    }
    break;
  case splashModeBGR8Packed:
    OPRS::error("Image: no supported color mode\n");
    result = splashErrOPVP;
    goto err0;
    break;
  }

  opvpctm.a = mat[0];
  opvpctm.b = mat[1];
  opvpctm.c = mat[2];
  opvpctm.d = mat[3];
  opvpctm.e = mat[4];
  opvpctm.f = mat[5];
  OPVP_i2Fix(0,opvprect.p0.x);
  OPVP_i2Fix(0,opvprect.p0.y);
  OPVP_i2Fix(1,opvprect.p1.x);
  OPVP_i2Fix(1,opvprect.p1.y);
  OPVP_i2Fix((tx),(opvpx));
  OPVP_i2Fix((ty),(opvpy));
  if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
    OPRS::error("SetCurrentPoint error\n");
  }

  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }
  if (opvp->DrawImage(printerContext,w,h,opvpcdepth,OPVP_iformatRaw,
       opvprect, opvpbytes*h, (void *)(buf)) < 0) {
    OPRS::error("DrawImage error\n");
    result = splashErrOPVP;
    goto err0;
  }
err0:
  /* reset CTM */
  opvpctm.a = 1.0;
  opvpctm.b = 0.0;
  opvpctm.c = 0.0;
  opvpctm.d = 1.0;
  opvpctm.e = 0.0;
  opvpctm.f = 0.0;
  if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
    OPRS::error("SetCTM error\n");
  }

  if (buf != 0) gfree(buf);
  return splashOk;
}

SplashError OPVPSplash::drawImage(SplashImageSource src, void *srcData,
			      SplashColorMode srcMode,
			      int w, int h, SplashCoord *mat,
			      GBool mask) {
  GBool ok, rot, halftone;
  SplashCoord xScale, yScale, xShear, yShear;
  int tx, ty, scaledWidth, scaledHeight, xSign, ySign;
  int ulx, uly, llx, lly, urx, ury, lrx, lry;
  int ulx1, uly1, llx1, lly1, urx1, ury1, lrx1, lry1;
  int xMin, xMax, yMin, yMax;
  SplashClipResult clipRes;
  SplashColor *pixBuf, *p;
  SplashColor pix;
  Guchar *alphaBuf, *q;
#ifdef OLD_DRAW_IMAGE
  SplashClipResult clipRes2;
  int yp, yq, yt, yStep, lastYStep;
  int xp, xq, xt, xStep, xSrc;
  int k1, spanXMin, spanXMax, spanY;
  SplashCoord pixAcc[splashMaxColorComps];
  int alphaAcc;
  SplashCoord pixMul, alphaMul, alpha;
  int x1, y1, x2, y2;
  int n, m;
#endif
  int x, y;
  int i, j;

  if (debugMode) {
    printf("drawImage: srcMode=%d w=%d h=%d mat=[%.2f %.2f %.2f %.2f %.2f %.2f]\n",
	   srcMode, w, h, mat[0], mat[1], mat[2], mat[3], mat[4], mat[5]);
  }

  // check color modes
  ok = gFalse; // make gcc happy
  switch (colorMode) {
  case splashModeMono1:
    ok = srcMode == splashModeMono1 || srcMode == splashModeMono8;
    break;
  case splashModeMono8:
    ok = srcMode == splashModeMono8;
    break;
  case splashModeRGB8:
    ok = srcMode == splashModeRGB8;
    break;
  case splashModeBGR8Packed:
    ok = srcMode == splashModeBGR8Packed;
    break;
  }
  if (!ok) {
    OPRS::error("Image Mode mismatch\n");
    return splashErrModeMismatch;
  }
  halftone = colorMode == splashModeMono1 && srcMode == splashModeMono8;

  // check for singular matrix
  if (splashAbs(mat[0] * mat[3] - mat[1] * mat[2]) < 0.000001) {
    OPRS::error("Image Not Singular Matrix\n");
    return splashErrSingularMatrix;
  }

  // compute scale, shear, rotation, translation parameters
  rot = splashAbs(mat[1]) > splashAbs(mat[0]);
  if (rot) {
    xScale = -mat[1];
    yScale = mat[2] - (mat[0] * mat[3]) / mat[1];
    xShear = -mat[3] / yScale;
    yShear = -mat[0] / mat[1];
  } else {
    xScale = mat[0];
    yScale = mat[3] - (mat[1] * mat[2]) / mat[0];
    xShear = mat[2] / yScale;
    yShear = mat[1] / mat[0];
  }
  tx = splashRound(mat[4]);
  ty = splashRound(mat[5]);
  scaledWidth = abs(splashRound(mat[4] + xScale) - tx) + 1;
  scaledHeight = abs(splashRound(mat[5] + yScale) - ty) + 1;
  xSign = (xScale < 0) ? -1 : 1;
  ySign = (yScale < 0) ? -1 : 1;

  // clipping
  ulx1 = 0;
  uly1 = 0;
  urx1 = xSign * (scaledWidth - 1);
  ury1 = splashRound(yShear * urx1);
  llx1 = splashRound(xShear * ySign * (scaledHeight - 1));
  lly1 = ySign * (scaledHeight - 1) + splashRound(yShear * llx1);
  lrx1 = xSign * (scaledWidth - 1) +
           splashRound(xShear * ySign * (scaledHeight - 1));
  lry1 = ySign * (scaledHeight - 1) + splashRound(yShear * lrx1);
  if (rot) {
    ulx = tx + uly1;    uly = ty - ulx1;
    urx = tx + ury1;    ury = ty - urx1;
    llx = tx + lly1;    lly = ty - llx1;
    lrx = tx + lry1;    lry = ty - lrx1;
  } else {
    ulx = tx + ulx1;    uly = ty + uly1;
    urx = tx + urx1;    ury = ty + ury1;
    llx = tx + llx1;    lly = ty + lly1;
    lrx = tx + lrx1;    lry = ty + lry1;
  }
  xMin = (ulx < urx) ? (ulx < llx) ? (ulx < lrx) ? ulx : lrx
                                   : (llx < lrx) ? llx : lrx
		     : (urx < llx) ? (urx < lrx) ? urx : lrx
                                   : (llx < lrx) ? llx : lrx;
  xMax = (ulx > urx) ? (ulx > llx) ? (ulx > lrx) ? ulx : lrx
                                   : (llx > lrx) ? llx : lrx
		     : (urx > llx) ? (urx > lrx) ? urx : lrx
                                   : (llx > lrx) ? llx : lrx;
  yMin = (uly < ury) ? (uly < lly) ? (uly < lry) ? uly : lry
                                   : (lly < lry) ? lly : lry
		     : (ury < lly) ? (ury < lry) ? ury : lry
                                   : (lly < lry) ? lly : lry;
  yMax = (uly > ury) ? (uly > lly) ? (uly > lry) ? uly : lry
                                   : (lly > lry) ? lly : lry
		     : (ury > lly) ? (ury > lry) ? ury : lry
                                   : (lly > lry) ? lly : lry;
  if ((clipRes = state->clip->testRect(xMin, yMin, xMax, yMax))
      == splashClipAllOutside) {
    return splashOk;
  }

  if (!noClipPath || clipRes == splashClipAllInside) {
    if (!mask && !noShearImage) {
      if (drawImageFastWithCTM(src,srcData,w,h,tx,ty,mat) == splashOk) {
	  return splashOk;
      }
    }
    if (!mask && splashRound(xShear) == 0 && splashRound(yShear) == 0) {
      /* no sheared case */
      if (drawImageNotShear(src,srcData,w,h,tx,ty,
	   scaledWidth, scaledHeight,xSign,ySign,rot) == splashOk) {
	  return splashOk;
      }
    }
  }

  /* shear case */
#ifdef OLD_DRAW_IMAGE
  /* display by pixel */


  // compute Bresenham parameters for x and y scaling
  yp = h / scaledHeight;
  yq = h % scaledHeight;
  xp = w / scaledWidth;
  xq = w % scaledWidth;

  // allocate pixel buffer
  pixBuf = (SplashColor *)gmalloc((yp + 1) * w * sizeof(SplashColor));
  alphaBuf = (Guchar *)gmalloc((yp + 1) * w * sizeof(Guchar));

  // init y scale Bresenham
  yt = 0;
  lastYStep = 1;

  for (y = 0; y < scaledHeight; ++y) {

    // y scale Bresenham
    yStep = yp;
    yt += yq;
    if (yt >= scaledHeight) {
      yt -= scaledHeight;
      ++yStep;
    }

    // read row(s) from image
    n = (yp > 0) ? yStep : lastYStep;
    if (n > 0) {
      p = pixBuf;
      q = alphaBuf;
      for (i = 0; i < n; ++i) {
	for (j = 0; j < w; ++j) {
	  (*src)(srcData, p++, q++);
	}
      }
    }
    lastYStep = yStep;

    // loop-invariant constants
    k1 = splashRound(xShear * ySign * y);

    // clipping test
    if (clipRes != splashClipAllInside &&
	!rot &&
	splashRound(yShear * k1) ==
	  splashRound(yShear * (xSign * (scaledWidth - 1) + k1))) {
      if (xSign > 0) {
	spanXMin = tx + k1;
	spanXMax = spanXMin + (scaledWidth - 1);
      } else {
	spanXMax = tx + k1;
	spanXMin = spanXMax - (scaledWidth - 1);
      }
      spanY = ty + ySign * y + splashRound(xShear * ySign * y);
      clipRes2 = state->clip->testSpan(spanXMin, spanXMax, spanY);
      if (clipRes2 == splashClipAllOutside) {
	continue;
      }
    } else {
      clipRes2 = clipRes;
    }

    // init x scale Bresenham
    xt = 0;
    xSrc = 0;

    for (x = 0; x < scaledWidth; ++x) {

      // x scale Bresenham
      xStep = xp;
      xt += xq;
      if (xt >= scaledWidth) {
	xt -= scaledWidth;
	++xStep;
      }

      // x shear
      x1 = xSign * x + k1;

      // y shear
      y1 = ySign * y + splashRound(yShear * x1);

      // rotation
      if (rot) {
	x2 = y1;
	y2 = -x1;
      } else {
	x2 = x1;
	y2 = y1;
      }

      // compute the filtered pixel at (x,y) after the x and y scaling
      // operations
      n = yStep > 0 ? yStep : 1;
      m = xStep > 0 ? xStep : 1;
      p = pixBuf + xSrc;
      q = alphaBuf + xSrc;
      for (i = 0; i < splashMaxColorComps; ++i) {
	pixAcc[i] = 0;
      }
      alphaAcc = 0;
      for (i = 0; i < n; ++i) {
	for (j = 0; j < m; ++j) {
	  switch (srcMode) {
	  case splashModeMono1:
	    pixAcc[0] += p->mono1;
	    break;
	  case splashModeMono8:
	    pixAcc[0] += p->mono8;
	    break;
	  case splashModeRGB8:
	    pixAcc[0] += splashRGB8R(p->rgb8);
	    pixAcc[1] += splashRGB8G(p->rgb8);
	    pixAcc[2] += splashRGB8B(p->rgb8);
	    break;
	  case splashModeBGR8Packed:
	    pixAcc[0] += splashBGR8R(p->bgr8);
	    pixAcc[1] += splashBGR8G(p->bgr8);
	    pixAcc[2] += splashBGR8B(p->bgr8);
	    break;
	  }
	  ++p;
	  alphaAcc += *q++;
	}
	p += w - m;
	q += w - m;
      }
      alphaMul = 1 / (SplashCoord)(n * m);
      if (halftone) {
	pixMul = (SplashCoord)alphaMul / 256.0;
      } else {
	pixMul = alphaMul;
      }
      alpha = (SplashCoord)alphaAcc * alphaMul;

      //~ this should blend if 0 < alpha < 1
      if (alpha > 0.75) {

	// mono8 -> mono1 conversion, with halftoning
	if (halftone) {
	  pix.mono1 = state->screen->test(tx + x2, ty + y2,
					  pixAcc[0] * pixMul);

	// no conversion, no halftoning
	} else {
          switch (colorMode) {
          case splashModeMono1:
            pix.mono1 = splashRound(pixAcc[0] * pixMul);
            break;
          case splashModeMono8:
            pix.mono8 = splashRound(pixAcc[0] * pixMul);
            break;
          case splashModeRGB8:
            pix.rgb8 = splashMakeRGB8(splashRound(pixAcc[0] * pixMul),
                                      splashRound(pixAcc[1] * pixMul),
                                      splashRound(pixAcc[2] * pixMul));
            break;
          case splashModeBGR8Packed:
            pix.bgr8 = splashMakeBGR8(splashRound(pixAcc[0] * pixMul),
                                      splashRound(pixAcc[1] * pixMul),
                                      splashRound(pixAcc[2] * pixMul));
            break;
          }
	}

	drawPixel(tx + x2, ty + y2, &pix, clipRes2 == splashClipAllInside);
      }

      // x scale Bresenham
      xSrc += xStep;
    }
  }

  gfree(pixBuf);
  gfree(alphaBuf);

  return splashOk;
#else
  SplashError result = splashOk;

  /* calculate inverse matrix */
  SplashCoord imat[4];
  double det = mat[0] * mat[3] - mat[1] * mat[2];
  imat[0] = mat[3]/det;
  imat[1] = -mat[1]/det;
  imat[2] = -mat[2]/det;
  imat[3] = mat[0]/det;

  /* read source image */
  pixBuf = (SplashColor *)gmalloc(h * w * sizeof(SplashColor));
  alphaBuf = (Guchar *)gmalloc(h * w * sizeof(Guchar));

  p = pixBuf;
  q = alphaBuf;
  for (i = 0; i < h; ++i) {
    for (j = 0; j < w; ++j) {
      (*src)(srcData, p++, q++);
    }
  }
  OPVP_Fix opvpx,opvpy;
  int opvpbytes, linesize;
  int opvpcdepth;
  OPVP_Rectangle opvprect;
  int width = xMax-xMin+1;
  int height = yMax-yMin+1;
  OPVP_CTM opvpctm;

  switch (colorMode) {
  case splashModeMono1:
  case splashModeMono8:
    opvpcdepth = 8;
    linesize = width;
    opvpbytes = (width+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeRGB8:
    opvpcdepth = 24;
    linesize = width*3;
    opvpbytes = (width*3+3)/4;
    opvpbytes *= 4;
    break;
  case splashModeBGR8Packed:
    OPRS::error("Image: no supported color mode\n");
    return splashErrOPVP;
    break;
  }
  /* allocate line buffer */
  Guchar *lineBuf = (Guchar *)gmalloc(opvpbytes);
  Guchar *onBuf = (Guchar *)gmalloc(width);
  SplashClip *clip;
  opvpctm.a = 1.0;
  opvpctm.b = 0.0;
  opvpctm.c = 0.0;
  opvpctm.d = 1.0;
  OPVP_i2Fix(0,opvprect.p0.x);
  OPVP_i2Fix(0,opvprect.p0.y);

  clip = state->clip->copy();
  if (w < scaledWidth || h < scaledHeight) {
    SplashPath cpath;

    cpath.moveTo(tx,ty);
    cpath.lineTo(mat[0]+tx,mat[1]+ty);
    cpath.lineTo(mat[0]+mat[2]+tx,mat[1]+mat[3]+ty);
    cpath.lineTo(mat[2]+tx,mat[3]+ty);
    clip->clipToPath(&cpath,1.0,gFalse);
  }
  for (y = 0;y < height;y++) {
    int dy = y+yMin-ty;
    memset(onBuf,0,width);
    for (x = 0;x < width;x++) {
      if (!clip->test(x+xMin,y+yMin)) continue;
      int ox,oy;
      /* calculate original coordinate */
      int dx = x+xMin-tx;
      ox = (int)trunc((imat[0]*dx+imat[2]*dy)*w);
      oy = (int)trunc((imat[1]*dx+imat[3]*dy)*h);
      if (ox >= 0 && ox < w && oy >= 0 && oy < h) {
	/* in the image */

	if (alphaBuf[oy*w+ox] != 0) {
	  switch (colorMode) {
	  case splashModeMono1:
	  case splashModeMono8:
	    lineBuf[x] = pixBuf[oy*w+ox].mono8;
	    break;
	  case splashModeRGB8:
	    pix = pixBuf[oy*w+ox];
	    lineBuf[x*3] = splashRGB8R(pix.rgb8);
	    lineBuf[x*3+1] = splashRGB8G(pix.rgb8);
	    lineBuf[x*3+2] = splashRGB8B(pix.rgb8);
	    break;
	  case splashModeBGR8Packed:
	    OPRS::error("Image: no supported color mode\n");
	    result = splashErrOPVP;
	    goto err1;
	    break;
	  }
	  onBuf[x] = 1;
	}
      }
    }
    /* out pixel */
    int sx = 0;
    int ex;
    while (sx < width) {
      /* find start pixel */
      for (;onBuf[sx] == 0 && sx < width;sx++);
      if (sx >= width) break;
      /* find end pixel */
      for (ex = sx+1;onBuf[ex] != 0 && ex < width;ex++);
      int n = ex-sx;
      Guchar *bp;
      int ns;

      switch (colorMode) {
      case splashModeMono1:
      case splashModeMono8:
	bp = lineBuf+sx;
	ns = n;
	break;
      case splashModeRGB8:
	bp = lineBuf+sx*3;
	ns = n*3;
	break;
      case splashModeBGR8Packed:
	OPRS::error("Image: no supported color mode\n");
	result = splashErrOPVP;
	goto err1;
	break;
      }
      ns = (ns+3)/4;
      ns *= 4;

      OPVP_i2Fix(xMin+sx,(opvpx));
      OPVP_i2Fix(yMin+y,(opvpy));
      if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
	OPRS::error("SetCurrentPoint error\n");
	result = splashErrOPVP;
	goto err1;
      }
      OPVP_i2Fix(n,opvprect.p1.x);
      OPVP_i2Fix(1,opvprect.p1.y);
      /* canonlisp driver use CTM only, ignores currentPoint */
      /* So, set start point to CTM */
      opvpctm.e = xMin+sx;
      opvpctm.f = yMin+y;
      if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
	OPRS::error("SetCTM error\n");
      }

      if (opvp->DrawImage(printerContext,n,1,opvpcdepth,OPVP_iformatRaw,
	   opvprect,ns, (void *)(bp)) < 0) {
	OPRS::error("DrawImage error\n");
	result = splashErrOPVP;
	goto err1;
      }

      /* reset CTM */
      opvpctm.e = 0.0;
      opvpctm.f = 0.0;
      if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
	OPRS::error("SetCTM error\n");
      }

      sx = ex+1;
    }
  }


err1:
  delete clip;
  gfree(pixBuf);
  gfree(lineBuf);
  gfree(onBuf);
  gfree(alphaBuf);
  return result;

#endif
}

void OPVPSplash::setColorMode(int colorModeA)
{
    colorMode = colorModeA;
}

void OPVPSplash::drawSpan(int x0, int x1, int y, GBool noClip)
{
  int s,e;
  OPVP_Point points[1];
  OPVP_Fix opvpx, opvpy;
  SplashCoord *savedLineDash = 0;
  int savedLineDashLength;
  SplashCoord savedLineDashPhase;
  SplashCoord savedLineWidth;
  GBool noSpan;


  if (opvp->NewPath(printerContext) < 0) {
    OPRS::error("NewPath error\n");
    return;
  }
  if (noClip) {
    noSpan = gFalse;
    OPVP_i2Fix(x0,opvpx);
    OPVP_i2Fix(y,opvpy);
    if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
      OPRS::error("SetCurrentPoint error\n");
      return;
    }
    OPVP_i2Fix(x1+1,points[0].x);
    OPVP_i2Fix(y,points[0].y);
    if (opvp->LinePath(printerContext,OPVP_PathOpen,1,points) < 0) {
      OPRS::error("LinePath error\n");
      return;
    }
  } else {
    noSpan = gTrue;
    s = x0;
    while (s < x1) {
      /* find start point */
      for (;s < x1;s++) {
	if (state->clip->test(s, y)) break;
      }
      if (s < x1) {
	/* start point was found */
	/* then find end point */
	for (e = s+1;e < x1;e++) {
	  if (!state->clip->test(e, y)) break;
	}
	/* do make span */
	noSpan = gFalse;
	OPVP_i2Fix(s,opvpx);
	OPVP_i2Fix(y,opvpy);
	if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
	  OPRS::error("SetCurrentPoint error\n");
	  return;
	}
	OPVP_i2Fix(e,points[0].x);
	OPVP_i2Fix(y,points[0].y);
	if (opvp->LinePath(printerContext,OPVP_PathOpen,1,points) < 0) {
	  OPRS::error("LinePath error\n");
	  return;
	}
	s = e;
      }
    }
  }
  if (opvp->EndPath(printerContext) < 0) {
    OPRS::error("EndPath error\n");
    return;
  }
  if (noSpan) return;
  /* change lins style temporarily */
  savedLineDashLength = state->lineDashLength;
  savedLineDashPhase = state->lineDashPhase;
  if (savedLineDashLength > 0 && state->lineDash != 0) {
    savedLineDash = new SplashCoord[savedLineDashLength];
    memcpy(savedLineDash, state->lineDash,
      savedLineDashLength*sizeof(SplashCoord));
  }
  setLineDash(0,0,0);
  savedLineWidth = state->lineWidth;
  setLineWidth(0.0);

  if (opvp->StrokePath(printerContext) < 0) {
    OPRS::error("StrokePath error\n");
    return;
  }

  /* restore line style */
  setLineDash(savedLineDash,savedLineDashLength,
    savedLineDashPhase);
  if (savedLineDash != 0) {
    delete[] savedLineDash;
  }
  setLineWidth(savedLineWidth);
}

/*
  draw pixel with StrokePath
  color is stroke color
*/
void OPVPSplash::drawPixel(int x, int y, GBool noClip)
{
  OPVP_Point points[1];
  OPVP_Fix opvpx, opvpy;

  if (noClip || state->clip->test(x, y)) {
    if (opvp->NewPath(printerContext) < 0) {
      OPRS::error("NewPath error\n");
      return;
    }
    OPVP_i2Fix(x,opvpx);
    OPVP_i2Fix(y,opvpy);
    if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
      OPRS::error("NewPath error\n");
      return;
    }
    OPVP_i2Fix(x+1,points[0].x);
    OPVP_i2Fix(y,points[0].y);
    if (opvp->LinePath(printerContext,OPVP_PathOpen,1,points) < 0) {
      OPRS::error("LinePath error\n");
      return;
    }
    if (opvp->EndPath(printerContext) < 0) {
      OPRS::error("EndPath error\n");
      return;
    }
    if (opvp->StrokePath(printerContext) < 0) {
      OPRS::error("StrokePath error\n");
      return;
    }
  }
}

#ifdef OLD_DRAW_IMAGE
/*
  draw pixel with DrawImage
*/
void OPVPSplash::drawPixel(int x, int y, SplashColor *color, GBool noClip)
{
  OPVP_Fix opvpx, opvpy;
  int opvpbytes;
  int opvpcdepth;
  OPVP_Rectangle opvprect;
  unsigned char opvpdata[4];
  OPVP_CTM opvpctm;

  if (noClip || state->clip->test(x, y)) {
    switch (colorMode) {
    case splashModeMono1:
      opvpdata[0] = splashRound(color->mono1);
      opvpcdepth = 1;
      opvpbytes = 1;
      break;
    case splashModeMono8:
      opvpdata[0] = splashRound(color->mono8);
      opvpcdepth = 8;
      opvpbytes = 1;
      break;
    case splashModeRGB8:
      opvpdata[0] = splashRound(splashRGB8R(color->rgb8));
      opvpdata[1] = splashRound(splashRGB8G(color->rgb8));
      opvpdata[2] = splashRound(splashRGB8B(color->rgb8));
      opvpdata[3] = 0;
      opvpcdepth = 24;
      opvpbytes = 4;
      break;
    case splashModeBGR8Packed:
     /* error */
      break;
    }
    OPVP_i2Fix((x),(opvpx));
    OPVP_i2Fix((y),(opvpy));
    if (opvp->SetCurrentPoint(printerContext,opvpx,opvpy) < 0) {
	OPRS::error("SetCurrentPoint error\n");
	return;
    }
    /* canonlisp driver use CTM only, ignores currentPoint */
    /* So, set start point to CTM */
    opvpctm.a = 1.0;
    opvpctm.b = 0.0;
    opvpctm.c = 0.0;
    opvpctm.d = 1.0;
    opvpctm.e = x;
    opvpctm.f = y;
    if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
      OPRS::error("SetCTM error\n");
    }

    OPVP_i2Fix(0,opvprect.p0.x);
    OPVP_i2Fix(0,opvprect.p0.y);
    OPVP_i2Fix(1,opvprect.p1.x);
    OPVP_i2Fix(1,opvprect.p1.y);
    if (opvp->DrawImage(printerContext,1,1,opvpcdepth,OPVP_iformatRaw,
	 opvprect,opvpbytes,(void *)opvpdata) < 0) {
	OPRS::error("DrawImage error\n");
    }

    /* reset CTM */
    opvpctm.e = 0.0;
    opvpctm.f = 0.0;
    if (opvp->SetCTM(printerContext,&opvpctm) < 0) {
      OPRS::error("SetCTM error\n");
    }
  }
}
#endif

char *OPVPSplash::getOption(char *key, int nOptions,
  char *optionKeys[], char *optionVals[])
{
  int i;

  for (i = 0;i < nOptions;i++) {
    if (strcmp(key,optionKeys[i]) == 0) {
      return optionVals[i];
    }
  }
  return 0;
}

void OPVPSplash::endPage()
{
  if (clipPath != 0) {
    delete clipPath;
    clipPath = 0;
  }
}

OPVPClipPath *OPVPClipPath::stackTop = 0;

OPVPClipPath::OPVPClipPath(SplashPath *pathA, GBool eoA)
{
  path = pathA;
  eo = eoA;
  next = 0;
  saved = gFalse;
}

void OPVPClipPath::push()
{
  OPVPClipPath *p;

  p = stackTop;
  stackTop = copy();
  stackTop->next = p;
  saved = gTrue;
}

OPVPClipPath *OPVPClipPath::pop() {
  OPVPClipPath *p = stackTop;
  if (stackTop != 0) stackTop = stackTop->next;
  return p;
}

OPVPClipPath *OPVPClipPath::copy()
{
  OPVPClipPath *p;

  p = new OPVPClipPath(path->copy(),eo);
  p->saved = saved;
  return p;
}
