/*
  OPVPWrapper.h
*/

#ifndef OPVPWRAPPER_H
#define OPVPWRAPPER_H

#include "opvp_common.h"

class OPVPWrapper {
public:
    OPVPWrapper(OPVP_api_procs *procsA);
    ~OPVPWrapper();

    OPVP_api_procs *getProcs() { return procs; }

    int ClosePrinter(int printerContext);
    int StartJob(int printerContext, char *jobInfo);
    int EndJob(int printerContext);
    int StartDoc(int printerContext, char *docInfo);
    int EndDoc(int printerContext);
    int StartPage(int printerContext, char *pageInfo);
    int EndPage(int printerContext);
    int ResetCTM(int printerContext);
    int SetCTM(int printerContext, OPVP_CTM *pCTM);
    int GetCTM(int printerContext, OPVP_CTM *pCTM);
    int InitGS(int printerContext);
    int SaveGS(int printerContext);
    int RestoreGS(int printerContext);
    int QueryColorSpace(int printerContext, OPVP_ColorSpace *pcspace,
	int *pnum);
    int SetColorSpace(int printerContext, OPVP_ColorSpace cspace);
    int GetColorSpace(int printerContext, OPVP_ColorSpace *pcspace);
    int QueryROP(int printerContext, int *pnum, int *prop);
    int SetROP(int printerContext, int rop);
    int GetROP(int printerContext, int *prop);
    int SetFillMode(int printerContext, OPVP_FillMode fillmode);
    int GetFillMode(int printerContext, OPVP_FillMode *pfillmode);
    int SetAlphaConstant(int printerContext, float alpha);
    int GetAlphaConstant(int printerContext, float *palpha);
    int SetLineWidth(int printerContext, OPVP_Fix width);
    int GetLineWidth(int printerContext, OPVP_Fix *pwidth);
    int SetLineDash(int printerContext, OPVP_Fix pdash[], int num);
    int GetLineDash(int printerContext, OPVP_Fix pdash[], int *pnum);
    int SetLineDashOffset(int printerContext, OPVP_Fix offset);
    int GetLineDashOffset(int printerContext, OPVP_Fix *poffset);
    int SetLineStyle(int printerContext, OPVP_LineStyle linestyle);
    int GetLineStyle(int printerContext, OPVP_LineStyle *plinestyle);
    int SetLineCap(int printerContext, OPVP_LineCap linecap);
    int GetLineCap(int printerContext, OPVP_LineCap *plinecap);
    int SetLineJoin(int printerContext, OPVP_LineJoin linejoin);
    int GetLineJoin(int printerContext, OPVP_LineJoin *plinejoin);
    int SetMiterLimit(int printerContext, OPVP_Fix miterlimit);
    int GetMiterLimit(int printerContext, OPVP_Fix *pmiterlimit);
    int SetPaintMode(int printerContext, OPVP_PaintMode paintmode);
    int GetPaintMode(int printerContext, OPVP_PaintMode *ppaintmode);
    int SetStrokeColor(int printerContext, OPVP_Brush *brush);
    int SetFillColor(int printerContext, OPVP_Brush *brush);
    int SetBgColor(int printerContext, OPVP_Brush *brush);
    int NewPath(int printerContext);
    int EndPath(int printerContext);
    int StrokePath(int printerContext);
    int FillPath(int printerContext);
    int StrokeFillPath(int printerContext);
    int SetClipPath(int printerContext, OPVP_ClipRule clipRule);
    int SetCurrentPoint(int printerContext, OPVP_Fix x, OPVP_Fix y);
    int LinePath(int printerContext, int flag, int npoints,
	OPVP_Point *points);
    int PolygonPath(int printerContext, int npolygons, int *nvertexes,
	OPVP_Point *points);
    int RectanglePath(int printerContext, int nrectangles,
	OPVP_Rectangle *reclangles);
    int RoundRectanglePath(int printerContext, int nrectangles,
	OPVP_RoundRectangle *reclangles);
#if (_PDAPI_VERSION_MAJOR_ == 0 && _PDAPI_VERSION_MINOR_ < 2)
    int BezierPath(int printerContext, int *npoints,
	OPVP_Point *points);
#else
    int BezierPath(int printerContext, int npoints,
	OPVP_Point *points);
#endif
    int ArcPath(int printerContext, int kind, int dir, OPVP_Fix bbx0,
	OPVP_Fix bby0, OPVP_Fix bbx1, OPVP_Fix bby1, OPVP_Fix x0,
	OPVP_Fix y0, OPVP_Fix x1, OPVP_Fix y1);
    int DrawBitmapText(int printerContext, int width, int height,
	int pitch, void *fontdata);
    int DrawImage(int printerContext, int sourceWidth, int sourceHeight,
	int colorDepth, OPVP_ImageFormat imageFormat,
	OPVP_Rectangle destinationSize, int count, void *imagedata);
    int StartDrawImage(int printerContext, int sourceWidth,
	int sourceHeight, int colorDepth, OPVP_ImageFormat imageFormat,
	OPVP_Rectangle destinationSize);
    int TransferDrawImage(int printerContext, int count,
	void *imagedata);
    int EndDrawImage(int printerContext);
    int StartScanline(int printerContext, int yposition);
    int Scanline(int printerContext, int nscanpairs, int *scanpairs);
    int EndScanline(int printerContext);
    int StartRaster(int printerContext, int rasterWidth);
    int TransferRasterData(int printerContext, int count,
	unsigned char *data);
    int SkipRaster(int printerContext, int count);
    int EndRaster(int printerContext);
    int StartStream(int printerContext);
    int TransferStreamData(int printerContext, int count, void *data);
    int EndStream(int printerContext);
#if (_PDAPI_VERSION_MAJOR_ > 0 || _PDAPI_VERSION_MINOR_ >= 2)
    int QueryDeviceCapability(int printerContext, int queryflag,
       int buflen, char *infoBuf);
    int QueryDeviceInfo(int printerContext, int queryflag,
       int buflen, char *infoBuf);
    int ResetClipPath(int printerContext);
#endif


private:
	OPVP_api_procs *procs;
};

#endif
