//========================================================================
//
// SplashFTFontFile.cc
//
//========================================================================

#include <aconf.h>

#if HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include "gmem.h"
#include "SplashFTFontEngine.h"
#include "SplashFTFont.h"
#include "SplashFTFontFile.h"

//------------------------------------------------------------------------
// SplashFTFontFile
//------------------------------------------------------------------------

SplashFontFile *SplashFTFontFile::loadType1Font(SplashFTFontEngine *engineA,
						SplashFontFileID *idA,
						char *fileNameA,
						GBool deleteFileA,
						char **encA) {
  FT_Face faceA;
  Gushort *codeToGIDA;
  char *name;
  int i;

  if (FT_New_Face(engineA->lib, fileNameA, 0, &faceA)) {
    return NULL;
  }
  codeToGIDA = (Gushort *)gmalloc(256 * sizeof(int));
  for (i = 0; i < 256; ++i) {
    codeToGIDA[i] = 0;
    if ((name = encA[i])) {
      codeToGIDA[i] = (Gushort)FT_Get_Name_Index(faceA, name);
    }
  }

  return new SplashFTFontFile(engineA, idA, fileNameA, deleteFileA,
			      faceA, gFalse, codeToGIDA, 256);
}

SplashFontFile *SplashFTFontFile::loadCIDFont(SplashFTFontEngine *engineA,
					      SplashFontFileID *idA,
					      char *fileNameA,
					      GBool deleteFileA,
					      Gushort *codeToGIDA,
					      int codeToGIDLenA) {
  FT_Face faceA;

  if (FT_New_Face(engineA->lib, fileNameA, 0, &faceA)) {
    return NULL;
  }

  return new SplashFTFontFile(engineA, idA, fileNameA, deleteFileA,
			      faceA, gFalse, codeToGIDA, codeToGIDLenA);
}

SplashFontFile *SplashFTFontFile::loadTrueTypeFont(SplashFTFontEngine *engineA,
						   SplashFontFileID *idA,
						   char *fileNameA,
						   GBool deleteFileA,
						   Gushort *codeToGIDA,
#ifdef HAVE_FT_OTLAYOUT
						   int codeToGIDLenA,
						   int wModeA,
						   FT_ULong scriptTag) {
#else
						   int codeToGIDLenA) {
#endif
  FT_Face faceA;
  GBool modeUnicodeA = gFalse;
#ifdef HAVE_FT_OTLAYOUT
  FTL_Glyphs_Array gaInA = 0, gaOutA = 0;
#endif

  if (FT_New_Face(engineA->lib, fileNameA, 0, &faceA)) {
    return NULL;
  }

#if 1
  if (!codeToGIDLenA) {
    int i;
    modeUnicodeA = gTrue;

#ifdef HAVE_FT_OTLAYOUT
    FTL_FeaturesRequest request;
    FT_UInt script_index, feature_index;
    FTL_EngineType engine_type;
    FT_Memory memory = faceA->stream->memory;

    if (wModeA == 1 && (faceA->face_flags & FT_FACE_FLAG_GLYPH_SUBSTITUTION)
       && FTL_Query_EngineType(faceA, &engine_type) == 0
       && engine_type == FTL_OPENTYPE_ENGINE) {
      if (FTL_New_FeaturesRequest(faceA, &request )) {
	goto fall_through;
      }
      if (!OTL_FeaturesRequest_Find_Script((OTL_FeaturesRequest)request,
	     OT_TABLE_GSUB, scriptTag,&script_index)) {
	goto fall_through;
      }
      if (!OTL_FeaturesRequest_Find_Feature((OTL_FeaturesRequest)request,
	     OT_TABLE_GSUB, OT_FEATURE_VRT2, script_index, 0xffff,
	       /* default language system */ &feature_index)) {
	  if (!OTL_FeaturesRequest_Find_Feature((OTL_FeaturesRequest)request,
		 OT_TABLE_GSUB, OT_FEATURE_VERT, script_index, 0xffff,
		   /* default language system */ &feature_index)) {
	    goto fall_through;
	  }
      }
      OTL_FeaturesRequest_Add_Feature((OTL_FeaturesRequest)request,
	      OT_TABLE_GSUB,feature_index,OT_VERT);
      FTL_Activate_FeaturesRequest(request);
      if (FTL_New_Glyphs_Array(memory,&gaInA)) {
	goto fall_through;
      }
      if (FTL_New_Glyphs_Array(memory,&gaOutA)) {
        FTL_Done_Glyphs_Array(gaInA);
	gaInA = 0;
	goto fall_through;
      }
      if (FTL_Set_Glyphs_Array_Length(gaInA,1)) {
        FTL_Done_Glyphs_Array(gaInA);
        FTL_Done_Glyphs_Array(gaOutA);
	gaInA = gaOutA = 0;
	goto fall_through;
      }
    }
fall_through:
#endif

    for (i = 0; i < faceA->num_charmaps; ++i) {
      if ((faceA->charmaps[i]->platform_id == 3 &&
           faceA->charmaps[i]->encoding_id == 1) ||
          faceA->charmaps[i]->platform_id == 0) {
        break;
      }
    }
    if (i == faceA->num_charmaps) {
      i = 0;
    }
    if (i) FT_Set_Charmap(faceA, faceA->charmaps[i]);
  } 
#endif

  return new SplashFTFontFile(engineA, idA, fileNameA, deleteFileA,
#ifdef HAVE_FT_OTLAYOUT
			      faceA, modeUnicodeA, codeToGIDA,
			      codeToGIDLenA, gaInA, gaOutA, wModeA);
#else
			      faceA, modeUnicodeA, codeToGIDA, codeToGIDLenA);
#endif
}

SplashFTFontFile::SplashFTFontFile(SplashFTFontEngine *engineA,
				   SplashFontFileID *idA,
				   char *fileNameA, GBool deleteFileA,
				   FT_Face faceA, GBool modeUnicodeA,
#ifdef HAVE_FT_OTLAYOUT
				   Gushort *codeToGIDA,
				   int codeToGIDLenA,
				   FTL_Glyphs_Array gaInA, 
				   FTL_Glyphs_Array gaOutA,
				   int wModeA):
#else
				   Gushort *codeToGIDA, int codeToGIDLenA):
#endif
  SplashFontFile(idA, fileNameA, deleteFileA)
{
  engine = engineA;
  face = faceA;
  codeToGID = codeToGIDA;
  codeToGIDLen = codeToGIDLenA;
  modeUnicode = modeUnicodeA;
#ifdef HAVE_FT_OTLAYOUT
  wMode = wModeA;
  gaIn = gaInA;
  gaOut = gaOutA;
#endif
}

SplashFTFontFile::~SplashFTFontFile() {
  if (face) {
    FT_Done_Face(face);
  }
  if (codeToGID && codeToGID != CODETOGID_IDENTITY) {
    gfree(codeToGID);
  }
#ifdef HAVE_FT_OTLAYOUT
  if (gaIn != 0) {
    FTL_Done_Glyphs_Array(gaIn);
  }
  if (gaOut != 0) {
    FTL_Done_Glyphs_Array(gaOut);
  }
#endif
}

SplashFont *SplashFTFontFile::makeFont(SplashCoord *mat) {
  SplashFont *font;

  font = new SplashFTFont(this, mat);
  font->initCache();
  return font;
}

#endif // HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H
